from ezlab.parameters import MAPR_CORE_VERSION, MAPR_MEP_VERSION


def get_local_repo(repo: str, repo_url: str):
    return rf"""[local-{repo.lower()}]
name = Local {repo}
enabled = 1
gpgcheck = 0
baseurl = {repo_url.rstrip('/')}/\$releasever/{repo}/\$basearch/os
ui_repoid_vars = releasever basearch
priority=1
proxy=
"""


def get_insecure_registry(registry):
    return f"""
{
  "log-driver": "journald",
  "insecure-registries" : [ {registry} ]
}
"""


def get_epel_repo(epelurl):
    return f"""
[local-epel]
name = Local EPEL
enabled = 1
gpgcheck = 0
baseurl = {epelurl}
priority=1
proxy="""


def get_mapr_repo(repo):
    return f"""
[ezmeral]
name = Ezmeral Packages
enabled = 1
gpgcheck = 0
baseurl = {repo.rstrip('/')}/v{MAPR_CORE_VERSION}/redhat
priority=1
proxy=

[ezmeral-eep]
name = Ezmeral EEP Packages
enabled = 1
gpgcheck = 0
baseurl = {repo.rstrip('/')}/MEP/MEP-{MAPR_MEP_VERSION}/redhat
priority=1
proxy=

"""



"""
Data Fabric operations user interface
"""

import logging
import re
from nicegui import app, ui, run

from ezlab.ezmeral.ezdf import client_install, cluster_install, xcluster

from ezlab.parameters import DF, DFAAS
from ezinfra.vms import prepare
from ezlab.utils import get_doc, get_fqdn


logger = logging.getLogger("ezlab.ui.df")


@ui.refreshable
def add_remove_host(hosts: list):
    if hosts is None:
        hosts = []
    for host in hosts:
        with ui.row().classes("w-full items-center no-wrap"):
            ui.input(
                "Hostname",
                placeholder="Host shortnames, ie, ezvm1,ezvm2...",
            ).bind_value(host, "name")
            ui.input(
                "IP Address",
                placeholder="10.1.1.x",
            ).bind_value(host, "ip")


async def prepare_action():
    hosts = app.storage.general[DF]["hosts"]
    isDFaaS = app.storage.general[DF]["dfaas"]

    # deep check list[dict] for missing values (if any host's name or IP is missing)
    if not (len(hosts) > 0 and all([values for values in [all(host) for host in [h.values() for h in hosts]]])):
        ui.notify("Missing values", type="warning")
        return False

    logger.info("Starting configuration for DF")
    app.storage.user["busy"] = True
    for host in hosts:
        result = await run.io_bound(
            prepare,
            hostname=get_fqdn(host["ip"]),
            hostip=host["ip"],
            settings=dict(app.storage.general["config"]),
            addhosts=[h for h in hosts if h["name"] != host["name"]],
            add_to_noproxy=(
                re.split(":|/", app.storage.general[DF].get("maprlocalrepo").split("://")[1])[0]
                if app.storage.general[DF].get("maprrepoislocal", False)
                else None
            ),
            prepare_for=DFAAS if isDFaaS else DF,
            dryrun=app.storage.general["config"]["dryrun"],
        )
        if app.storage.general["config"]["dryrun"]:
            get_doc(result)
        elif result:
            logger.info("[ %s ] ready: %s", host["name"], result)
        else:
            ui.notify(f"[{host}] preparation failed!", type="warning")

    app.storage.user["busy"] = False


def prepare_menu():
    if not "hosts" in app.storage.general[DF].keys():
        app.storage.general[DF]["hosts"] = [{"name": "", "ip": ""}]

    ui.checkbox("DF as a Service?").bind_value(app.storage.general[DF], "dfaas")

    with ui.row().classes("w-full items-center items-stretch") as prep_view:
        ui.label("Hosts")
        ui.space()
        with ui.row():

            def add_host():
                try:
                    app.storage.general[DF]["hosts"].append({"name": "", "ip": ""})
                except (KeyError, AttributeError):
                    app.storage.general[DF]["hosts"] = [{"name": "", "ip": ""}]
                finally:
                    add_remove_host.refresh(app.storage.general[DF]["hosts"])

            def remove_host():
                if len(app.storage.general[DF]["hosts"]) > 0:
                    app.storage.general[DF]["hosts"].pop()
                    add_remove_host.refresh()

            ui.button(icon="add_box", color=None, on_click=add_host).props("flat")
            ui.button(icon="delete", color=None, on_click=remove_host).props("flat")

    # Hosts
    add_remove_host(app.storage.general[DF]["hosts"])


async def install_action(hosts_string: str):

    logger.info("Starting DF cluster installation")

    app.storage.user["busy"] = True

    result = await run.io_bound(
        cluster_install,
        hosts=hosts_string.split(","),
        settings=dict(app.storage.general["config"]),
        config=dict(app.storage.general[DF]),
        dryrun=app.storage.general["config"]["dryrun"],
    )

    if app.storage.general["config"]["dryrun"]:
        get_doc(result)

    elif result:
        ui.notify(f"Cluster is ready!", type="positive")

    else:
        ui.notify(f"Cluster creation failed!", type="negative")

    app.storage.user["busy"] = False


async def xcluster_action():

    logger.info("Starting DF cross-cluster setup: DISABLED")

    app.storage.user["busy"] = True

    if not await run.io_bound(
        xcluster,
        config=dict(app.storage.general["config"]),
        settings=dict(app.storage.general[DF]),
        dryrun=app.storage.general["config"]["dryrun"],
    ):
        ui.notify(f"Cluster creation failed!", type="negative")
        return False

    app.storage.user["busy"] = False
    return True


async def client_action():

    logger.info("Starting DF client configuration")

    app.storage.user["busy"] = True

    result = await run.io_bound(
        client_install,
        username=app.storage.general["config"]["username"],
        keyfile=app.storage.general["config"]["privatekeyfile"],
        config=dict(app.storage.general[DF]),
    )

    if not result:
        ui.notify("Client configuration failed", type="negative")

    app.storage.user["busy"] = False

    if result:
        ui.notify(f"Finished client configuration", type="positive")
        return True
    else:
        return False



# /usr/bin/python3

from ipaddress import ip_address
import logging
import os
import string
import libvirt
from libvirt import *
from xml.etree import ElementTree as ET
from ezinfra.cloudinit import cloudinit_metadata, cloudinit_networkconfig, cloudinit_userdata
from ezinfra.sshkeys import get_privatekey_from_string, get_ssh_pubkey_from_privatekey
from ezinfra.remote import sftp_copy_file, wait_for_ssh
from ezlab.utils import create_ci_iso, ezapp

logger = logging.getLogger("ezinfra.kvm")


def connect(host: str, username: str):
    try:
        if host in ["localhost", "127.0.0.1"]:
            target = "qemu:///session"

        else:
            where = "session" if host == "localhost" else "system"
            target = f"qemu+ssh://{username}@{host}/{where}"

        conn = open(target)
    except libvirt.libvirtError as e:
        print(f"ERROR IN CONNECT: {e}")
        return e

    return conn


def vms():
    domainIDs = ezapp.connection.listDomainsID()
    if domainIDs == None or len(domainIDs) == 0:
        return list()
    else:
        for domainID in domainIDs:
            yield domainID


def bridges():
    return [n.bridgeName() for n in [ezapp.connection.networkLookupByName(net) for net in ezapp.connection.listNetworks()]]


def pools():
    pools = ezapp.connection.listAllStoragePools()
    if not pools:
        return list()
        # raise SystemExit("Failed to locate any StoragePool objects.")

    # return pools
    return [pool for pool in pools if pool.isActive() == True]


def volumes(pool):
    if pool is None:
        return list()

    sp = ezapp.connection.storagePoolLookupByName(pool)

    if sp is None:
        return list()
    else:
        return [vol for vol in sp.listVolumes() if ".qcow2" in vol or ".img" in vol]


def sp_basepath(sp):
    raw_xml = sp.XMLDesc(0)
    tree = ET.ElementTree(ET.fromstring(raw_xml))

    return tree.find("target").find("path").text


def get_storagepool(poolname):
    return ezapp.connection.storagePoolLookupByName(poolname)


def clone(
    resources: set,
    settings: dict,
    # vm_number: int,
    vm_count: int,
    dryrun=True,
):
    """
    Clone VM
    """

    response = {}
    response["settings"] = settings

    dom = None

    # Parse inputs
    (pool, baseimg, bridge, eznode, hostname, first_ip) = resources

    privatekey: str = settings["privatekey"]
    ssh_keyfile: str = settings["privatekeyfile"]

    pk = get_privatekey_from_string(privatekey)
    publickey = get_ssh_pubkey_from_privatekey(pk)
    # kvm_host = ezapp.connection.getHostname()
    _, host = ezapp.connection.getURI().split("://")
    kvm_username, kvm_target = host.split("@")
    kvm_host, _ = kvm_target.split("/")

    sp = get_storagepool(pool)
    # host path for storage pool
    sp_path = sp_basepath(sp)
    baseimgPath = sp.storageVolLookupByName(baseimg).path()

    for vm_number in range(vm_count):

        multivm = vm_count > 1
        vm_name = hostname + str(vm_number if multivm else "")
        vm_domain = settings["domain"]
        vm_ip = str(ip_address(first_ip) + vm_number if multivm else ip_address(first_ip))  # adjust for single VM
        vm_gateway = settings["gateway"]
        vm_network_bits = settings["cidr"].split("/")[1]

        if dryrun:
            logger.info("Would clone %s from %s", vm_name, baseimg)
        else:
            logger.info("[ %s ] cloning...", vm_name)

        # Cloud-init files
        cloudinit_files = {
            # meta-data
            "meta-data": cloudinit_metadata(hostname=vm_name),
            # user-data
            "user-data": cloudinit_userdata(
                hostname=vm_name,
                fqdn=f"{vm_name}.{vm_domain}",
                ssh_username=settings["username"],
                ssh_password=settings["password"],
                ssh_pubkey=publickey,
                swap_disk="/dev/vdb",
            ),
            # network-config
            "network-config": cloudinit_networkconfig(
                interface="ens1" if "ubuntu" in baseimg.lower() else "eth0",
                ip_address=vm_ip,
                network_bits=vm_network_bits,
                gateway=vm_gateway,
                nameserver=settings["nameserver"],
                searchdomain=vm_domain,
            ),
        }

        response[
            "task"
        ] += f"""
        Clone {vm_name}:
            OS Disk: {eznode["os_disk_size"]}G
            # Swap Disk: {eznode['swap_disk_size']}G
            Data Disks (qty: {eznode["no_of_disks"]}): {eznode['data_disk_size'] if 'data_disk_size' in eznode else 0}G
            Disk Location {kvm_host}:{sp_path}/{vm_name}/
            """
        response["files"] = cloudinit_files

        # copy cloudinit files
        if not dryrun:
            logger.info("[ %s ] uploading cloud-init files...", vm_name)

            # need absolute path for filename
            isofile = os.path.abspath(f"{vm_name}-ci.iso")

            try:
                create_ci_iso(files=cloudinit_files, destfile=isofile)

                sftp_copy_file(
                    host=kvm_host,
                    username=kvm_username,
                    keyfile=ssh_keyfile,
                    source=isofile,
                    destfile=f"{sp_path}/{vm_name}-ci.iso",
                )

            except Exception as error:
                logger.warning("Failed to copy cloudinit %s", error)
                return False

            finally:
                try:
                    os.remove(isofile)
                except OSError:
                    pass

            logger.info("[ %s ] creating disks...", vm_name)

            # OS disk, backed by image
            create_osvol(
                sp=sp,
                volname=f"{vm_name}-os.qcow2",
                sizeGB=eznode["os_disk_size"],
                baseimg=baseimgPath,
            )

            # # swap disk
            # create_datavol(
            #     sp=sp,
            #     volname=f"{vm_name}-swap.qcow2",
            #     sizeGB=eznode["swap_disk_size"],
            # )

            # data disk(s)
            for i in range(eznode["no_of_disks"]):
                create_datavol(
                    sp=sp,
                    volname=f"{vm_name}-data{i}.qcow2",
                    sizeGB=eznode["data_disk_size"],
                )

        # Define VM
        vmXml = vm_xml(
            name=vm_name,
            memMB=int(eznode["memGB"] * 1024),
            cores=eznode["cores"],
            basepath=sp_path,
            baseimgPath=baseimgPath,
            bridge=bridge,
            no_of_datadisks=int(eznode["no_of_disks"]),
        )

        if dryrun:
            response["commands"] = [
                f"""virsh dom create <<EOF
                    {vmXml}
                EOF
                """
            ]
        else:
            logger.info("[ %s ] creating vm...", vm_name)
            try:
                dom = ezapp.connection.defineXMLFlags(vmXml, 0)

            except Exception as error:
                logger.warning("Failed at VM definition %s: %s", vm_name, error)
                return False

        # Update dns entry in virtualnet
        if dryrun:
            response[f"commands"].append(f"echo '{vm_ip} {vm_name}.{vm_domain}, {vm_name}' >> /etc/hosts")
        else:
            # ### THIS IS WRONG, NIC SHOULD BE DETACHED/ATTACHED ON ALL RUNNING INSTANCES
            try:
                logger.info("[ %s ] Update libvirt dns", vm_name)
                virnet = [
                    ezapp.connection.networkLookupByName(net)
                    for net in ezapp.connection.listNetworks()
                    if ezapp.connection.networkLookupByName(net).bridgeName() == bridge
                ].pop()
                if virnet:
                    dnshostXml = f"<host ip='{vm_ip}'><hostname>{vm_name}</hostname></host>"
                    if (
                        virnet.update(
                            command=libvirt.VIR_NETWORK_UPDATE_COMMAND_ADD_LAST,
                            section=libvirt.VIR_NETWORK_SECTION_DNS_HOST,
                            parentIndex=-1,
                            xml=dnshostXml,
                            flags=libvirt.VIR_DOMAIN_AFFECT_CONFIG | libvirt.VIR_DOMAIN_AFFECT_LIVE,
                        )
                        == 0
                    ):
                        logger.info("[ %s ] dns entry added for %s", vm_name, vm_ip)
                    else:
                        logger.info("[ %s ] dns update failed for %s, ignoring", vm_name, vm_ip)
                else:
                    logger.info("[ %s ] dns update failed for %s, are you using a bridge?", vm_name, vm_ip)

            except Exception as error:
                logger.warning("[ %s ] Failed to update virtualNetwork: %s, ignoring...", vm_name, error)

        # Create/Boot VM
        if not dryrun:
            logger.info("[ %s ] starting...", vm_name)
            try:
                if dom.create() < 0:
                    logger.warning("[ %s ] failed to create: %s", vm_name, error)
                    return False

                logger.info("[ %s ] booting", vm_name)
                dom.setAutostart(1)  # turn on autostart
            except Exception as error:
                logger.warning("[ %s ] failed to boot: %s", vm_name, error)
                return False

            logger.info("[ %s ] waiting startup...", vm_name)
            if not wait_for_ssh(f"{vm_name}.{vm_domain}", settings["username"], ssh_keyfile):
                logger.warning("[ %s ] ssh failed", dom.name())
                return False

            logger.info("[ %s ] ready for %s", dom.name(), eznode["product"])

    return response if dryrun else True


def vm_xml(name: str, memMB: int, cores: int, basepath: str, baseimgPath: str, bridge: str, no_of_datadisks: int):

    datadisks = ""
    letters = string.ascii_lowercase
    start_letter = letters.index("c")  # starting from /dev/vdc

    for i in range(no_of_datadisks):
        diskname = f"vd{letters[start_letter + i]}"
        datadisks += f"""
    <disk type='file' device='disk'>
      <driver name='qemu' type='qcow2' discard='unmap'/>
      <source file='{basepath}/{name}-data{i}.qcow2'/>
      <backingStore/>
      <target dev='{diskname}' bus='virtio'/>
      <alias name='virtio-disk{i+2}'/>
    </disk>
"""

    return f"""
<domain type="kvm">
  <name>{name}</name>
  <metadata>
    <libosinfo:libosinfo xmlns:libosinfo="http://libosinfo.org/xmlns/libvirt/domain/1.0">
      <libosinfo:os id="http://rockylinux.org/rocky/8"/>
    </libosinfo:libosinfo>
  </metadata>
  <memory unit='MiB'>{memMB}</memory>
  <currentMemory unit='MiB'>{memMB}</currentMemory>
  <memoryBacking>
    <source type='memfd'/>
    <access mode='shared'/>
  </memoryBacking>
  <vcpu placement='static'>{cores}</vcpu>
  <features>
    <acpi/>
    <apic/>
    <vmport state='off'/>
  </features>
  <resource>
    <partition>/machine</partition>
  </resource>
  <sysinfo type='smbios'>
    <system>
      <entry name='serial'>ds=nocloud</entry>
    </system>
  </sysinfo>
  <os>
    <type arch='x86_64' machine='pc-q35-6.2'>hvm</type>
    <boot dev='hd'/>
    <smbios mode='sysinfo'/>
  </os>
  <devices>
    <clock offset="utc"/>
    <on_poweroff>destroy</on_poweroff>
    <on_reboot>restart</on_reboot>
    <on_crash>destroy</on_crash>
    <disk type='file' device='disk'>
      <driver name='qemu' type='qcow2' discard='unmap'/>
      <source file='{basepath}/{name}-os.qcow2'/>
      <backingStore type='file'>
        <format type='qcow2'/>
        <source file='{baseimgPath}'/>
        <backingStore/>
      </backingStore>
      <target dev='vda' bus='virtio'/>
      <alias name='virtio-disk0'/>
    </disk>
    # <disk type='file' device='disk'>
    #   <driver name='qemu' type='qcow2' discard='unmap'/>
    #   <source file='{basepath}/{name}-swap.qcow2'/>
    #   <backingStore/>
    #   <target dev='vdb' bus='virtio'/>
    #   <alias name='virtio-disk1'/>
    # </disk>
    <disk type='file' device='cdrom'>
      <driver name='qemu' type='raw'/>
      <source file='{basepath}/{name}-ci.iso'/>
      <backingStore/>
      <target dev='sda' bus='sata'/>
      <readonly/>
      <alias name='sata0-0-0'/>
      <address type='drive' controller='0' bus='0' target='0' unit='0'/>
    </disk>
    {datadisks}
    <interface type="bridge">
      <source bridge="{bridge}"/>
    </interface>
    <input type="mouse" bus="ps2"/>
    <graphics type="vnc" autoport="yes" listen="127.0.0.1"/>
    <console type="pty">
      <source>
        <clipboard copypaste="yes"/>
      </source>
      <target type="virtio"/>
    </console>
    <memballoon model='virtio'/>
    <rng model='virtio'>
      <backend model='random'>/dev/urandom</backend>
      <alias name='rng0'/>
    </rng>
  </devices>
</domain>
"""


def create_datavol(sp: virStoragePool, volname: str, sizeGB: int):
    volXml = f"""
        <volume>
          <name>{volname}</name>
          <allocation>0</allocation>
          <capacity unit="GiB">{sizeGB}</capacity>
          <target>
            <format type='qcow2'/>
            <permissions>
              <mode>0644</mode>
              <label>virt_image_t</label>
            </permissions>
          </target>
        </volume>"""

    return sp.createXML(volXml, 0)


def create_osvol(sp: virStoragePool, volname: str, sizeGB: int, baseimg: str):
    volXml = f"""
        <volume>
          <name>{volname}</name>
          <allocation>0</allocation>
          <capacity unit="GiB">{sizeGB}</capacity>
          <target>
            <format type='qcow2'/>
            <backingStore type='file'>
                <format type='qcow2'/>
                <source file='{baseimg}'/>
                <backingStore/>
            </backingStore>
            <permissions>
              <mode>0644</mode>
              <label>virt_image_t</label>
            </permissions>
          </target>
        </volume>"""

    return sp.createXML(volXml, 0)


# def get_network_with_new_host(network: virNetwork, hostname: str, ip: str):

#     # <dns>
#     #   <host ip='192.168.122.2'>
#     #     <hostname>myhost</hostname>
#     #   </host>
#     # </dns>

#     raw_xml = network.XMLDesc(0)
#     tree = ET.ElementTree(ET.fromstring(raw_xml))
#     dnsDom = tree.find("dns")
#     if not dnsDom:
#         print("add dns entry")
#         dnsDom = ET.Element("dns")
#         tree.getroot().append(dnsDom)

#     else:
#         # check if ip exist (append host if so)
#         existing_ip = [h for h in dnsDom if h.get("ip") == ip]
#         if len(existing_ip) > 0:
#             hostElement = existing_ip[0]
#             if hostElement.find("hostname").text == hostname:
#                 # nothing to change
#                 return None
#             else:
#                 newhost = ET.Element("hostname")
#                 newhost.text = hostname
#                 hostElement.append(newhost)

#     # new host entry
#     newhost = ET.Element("hostname")
#     newhost.text = hostname
#     newip = ET.Element("host", ip=ip)
#     newip.append(newhost)
#     dnsDom.append(newip)

#     return ET.tostring(tree.getroot(), 'unicode')


# def get_nicXml(dom: virDomain, bridge: str):
#     tree = ET.ElementTree(
#         ET.fromstring(dom.XMLDesc(0))
#     )
#     root = tree.getroot()

#     interfaces = root.findall("./devices/interface")
#     nic = [nic for nic in interfaces if nic.find("source").attrib.get("bridge") == bridge].pop()

#     return ET.tostring(nic, 'unicode')


#         netXml = get_network_with_new_host(virnet, vm_name, vm_ip)
#         if netXml is not None:
#             virnet.destroy()
#             virnet.undefine()
#             network = ezapp.connection.networkDefineXML(netXml)
#             if network == None:
#                 print("Failed to create network", netXml)
#             else:
#                 network.create()
#                 network.setAutostart(1)
#                 # remove/attach VM nic
#                 nicXml = get_nicXml(dom, bridge)
#                 dom.dev
#                 dom.attachDeviceFlags(nicXml, libvirt.VIR_DOMAIN_AFFECT_LIVE)



from ipaddress import ip_address
import logging
from time import sleep
from urllib.parse import quote
from proxmoxer import ProxmoxAPI
from ezinfra.sshkeys import get_ssh_pubkey_from_privatekey, get_privatekey_from_string
from ezinfra.remote import ssh_run_command, wait_for_ssh
from ezlab.utils import ezapp, get_new_vm_name

logger = logging.getLogger("ezinfra.pve")


def connect(host, username, password):
    result = None
    try:
        result = ProxmoxAPI(
            host,
            user=username,
            password=password,
            verify_ssl=False,
        )
    except Exception as error:
        return error

    return result


def vms():
    return ezapp.connection.cluster.resources.get(type="vm")


def pools():
    return ezapp.connection.cluster.resources.get(type="storage")


def networks():
    return ezapp.connection.cluster.resources.get(type="sdn")


# def bridges(fromtemplatename):
#     if not fromtemplatename or fromtemplatename == "":
#         return []

#     try:
#         template = next(iter([t for t in vms() if t["template"] and t["name"] == fromtemplatename]))
#     except IndexError:
#         return []

#     return ezapp.connection.nodes(template["node"]).network.get(type="any_bridge") if template else []


def bridges(fromtemplateid):
    if not fromtemplateid or fromtemplateid == "":
        return []

    try:
        template = next(iter([t for t in vms() if t["template"] and t["id"] == fromtemplateid]))
    except IndexError:
        return []

    return ezapp.connection.nodes(template["node"]).network.get(type="any_bridge") if template else []


def clone(
    resources: tuple,
    settings: dict,
    # vm_number: int,
    vm_count: int,
    dryrun: bool,
):
    template, volume, bridge, eznode, hostname, first_ip = resources

    response = {}
    response["commands"] = []
    response["task"] = f"Clone {vm_count} VMs for {eznode['name']}"
    response["settings"] = settings
    response["info"] = ""
    response["files"] = {}

    node = template["node"]
    template_type = template["type"]
    template_id = template["vmid"]

    vm_gateway = settings["gateway"]
    vm_network_bits = settings["cidr"].split("/")[1]
    privatekey: str = settings["privatekey"]

    # cloudinit requires OpenSSH format public key
    pk = get_privatekey_from_string(privatekey)
    publickey = get_ssh_pubkey_from_privatekey(pk)

    for vm_number in range(vm_count):
        multivm = vm_count > 1
        if multivm:
            vm_name = get_new_vm_name(hostname, vm_number)
        else:
            vm_name = hostname

        response[
            "info"
        ] += f"""
        Clone {vm_name} with disks:
            OS Disk: {eznode["os_disk_size"]}G
            # Swap Disk: {eznode['swap_disk_size']}G
            Data Disks (qty: {eznode["no_of_disks"]}): {eznode['data_disk_size'] if 'data_disk_size' in eznode else 0}G
            """

        if not dryrun:
            logger.info("[ %s ] cloning...", vm_name)
            # wait for others to request vmid, used when run for multiple VMs in parallel
            sleep(vm_number * 1)
            nextid = ezapp.connection.cluster.nextid.get()
            logger.info("[ %s ] assigned id %s", vm_name, nextid)

        ipconfig = "ip="
        vm_ip = ""
        if first_ip == "dhcp":
            ipconfig += "dhcp"
        else:
            vm_ip = str(ip_address(first_ip) + vm_number if multivm else ip_address(first_ip))  # adjust for single VM
            ipconfig += f"{vm_ip}/{vm_network_bits}"
            ipconfig += f",gw={vm_gateway}"

        if not dryrun:
            logger.info("[ %s ] (%s) creating", vm_name, nextid)
            try:
                result = task_waitfor(
                    ezapp.connection,
                    node,
                    ezapp.connection.nodes(node)(template_type)(template_id).clone.post(
                        newid=nextid,
                        name=vm_name,
                        description=eznode["name"],
                    ),
                )
                if result:
                    new_vm = ezapp.connection.nodes(node)(template_type)(nextid)
                    logger.info("[ %s ] cloned", vm_name)
                else:
                    logger.warning("Clone failed for %s: %s", vm_name, result)
                    return False

            except Exception as error:
                logger.warning("PVE Exception for %s: %s", vm_name, error)
                return False

        if dryrun:
            response["files"].update(
                {
                    f"{vm_name}-ciconfig": f"""
                    cores={eznode["cores"]},
                    memory={eznode["memGB"] * 1024},
                    net0=virtio,bridge={bridge},firewall=0,
                    ipconfig0={ipconfig},
                    tags={eznode["product"]},
                    ciuser={settings["username"]},
                    cipassword=******,
                    nameserver={settings["nameserver"]},
                    searchdomain={settings["domain"]},
                    ciupgrade=0,
                    sshkeys={quote(publickey, safe="")},
                    onboot=1,
                    efidisk0={volume["storage"]}:1,efitype=4m,pre-enrolled-keys=1,size=4M",
                    """,
                }
            )

        else:
            # configure vm
            logger.info("[ %s ] reconfigure", vm_name)
            try:
                new_vm.config.post(
                    cores=eznode["cores"],
                    memory=eznode["memGB"] * 1024,
                    net0=f"virtio,bridge={bridge},firewall=0",
                    ipconfig0=ipconfig,
                    tags=eznode["product"],
                    ciuser=settings["username"],
                    cipassword=settings["password"],
                    nameserver=settings["nameserver"],
                    searchdomain=settings["domain"],
                    ciupgrade=0,
                    sshkeys=quote(publickey, safe=""),
                    onboot=1,
                    efidisk0=f"{volume['storage']}:1,efitype=4m,pre-enrolled-keys=1,size=4M",
                )
                logger.info("[ %s ] reconfigured", vm_name)
            except Exception as error:
                logger.warning("PVE Exception for %s: %s", vm_name, error)
                return False

        if not dryrun:
            logger.info("[ %s ] add disks", vm_name)
            try:
                # configure disks
                new_vm.resize.put(disk="scsi0", size=f"{eznode['os_disk_size']}G")
                # add swap disk
                # swap_disk = f"{volume['storage']}:{eznode['swap_disk_size']},backup=0,discard=on,cache=unsafe,iothread=1,replicate=0,ssd=1"
                # new_vm.config.post(scsi1=swap_disk)
                # logger.info("[ %s ] %dGB swap disk added", vm_name, eznode["swap_disk_size"])

                # add data disks
                data_disk = f"{volume['storage']}:{eznode['data_disk_size'] if 'data_disk_size' in eznode else 0},backup=0,discard=on,cache=unsafe,iothread=1,replicate=0,ssd=1"

                disks = {}
                for diskno in range(eznode["no_of_disks"]):
                    # adding 2 since 0 is OS, 1 is SWAP
                    disks.update({"scsi" + str(diskno + 2): data_disk})

                # add all data disks at once, using (scsi2=xxx, scsi3=xxx) as params
                new_vm.config.post(**disks)

                logger.info("[ %s ] disks attached", vm_name)

            except Exception as error:
                logger.warning("PVE Exception for %s: %s", vm_name, error)
                return False

            # start vm
            new_vm.status.start.post()

            logger.info("[ %s ] waiting startup...", vm_name)

            # # apply customisations to vm
            if not wait_for_ssh(vm_ip, settings["username"], settings["privatekey"]):
                logger.warning(f"[ %s ] SSH FAILED", vm_ip)
                return False

            # # Setup swap space
            # for out in ssh_run_command(
            #     host=vm_ip,
            #     username=settings["username"],
            #     keyfile=settings["privatekeyfile"],
            #     command="sudo mkswap /dev/sdb; sudo swapon /dev/sdb",
            # ):
            #     logger.info("[ %s ] swap on: %s", vm_name, out)
            #     UUID = out.split(" UUID=")[1]
            #     if UUID:
            #         for out in ssh_run_command(
            #             host=vm_ip,
            #             username=settings["username"],
            #             keyfile=settings["privatekeyfile"],
            #             command=f"echo 'UUID={UUID} none swap sw,nofail 0 0' | sudo tee -a /etc/fstab",
            #         ):
            #             logger.debug("add swap to fstab %s", out)

            # # reboot for changes
            # if task_waitfor(
            #     proxmox=ezapp.connection,
            #     node=node,
            #     task_id=new_vm.status.reboot.post(timeout=60),
            # ) == TASK_FINISHED:
            #     sleep(30) # allow VM to reboot
            #     logger.info("[ %s ] ready for %s", vm_name, eznode["product"])
            #     return True

            logger.info("[ %s ] Finished", vm_name)

    return response if dryrun else True


def task_waitfor(proxmox, node, task_id):
    task_submitted = proxmox.nodes(node).tasks(task_id).status.get()
    if task_submitted["status"] == "stopped":
        return task_submitted["exitstatus"]

    try:
        status = proxmox.nodes(node).tasks(task_id).status.get()
        while status["status"] != "stopped":
            logger.debug("PVE task status: %s is %s", status["type"], status["status"])
            sleep(3)
            status = proxmox.nodes(node).tasks(task_id).status.get()
    except Exception as error:
        logger.warning("TASK WAIT FAILED: %s", error)
        return False

    return True



"""
DESCRIPTION:
    Simple CLI tool to deploy my lab VMs
USAGE EXAMPLE:
    > ezlabctl create template|vm --target pve --host <target-host>
    > ezlabctl ezdf cluster host1 host2 host3 --username --token
"""

# ::IMPORTS ------------------------------------------------------------------------ #

# cli framework - https://pypi.org/project/typer/
from enum import Enum
import os
from typing import Annotated, Optional
import typer
from ezlab.parameters import DF, GENERIC, UA

# package for reading details about this package
from ezlabctl.common import prepare_vm

from ezlabctl import config
from ezlabctl import create
from ezlabctl import delete
from ezlabctl import ezua
from ezlabctl import ezdf
from ezlabctl.parameters import ini_file

app = typer.Typer(add_completion=True, no_args_is_help=True)

# ::SETUP -------------------------------------------------------------------------- #
app = typer.Typer(add_completion=True, no_args_is_help=True)

# ::SETUP SUBPARSERS --------------------------------------------------------------- #
app.add_typer(
    config.app,
    no_args_is_help=True,
    name="config",
    short_help="manage deployment settings",
)
app.add_typer(
    create.app,
    no_args_is_help=True,
    name="create",
    short_help="create templates and VMs",
)
app.add_typer(
    delete.app,
    no_args_is_help=True,
    name="delete",
    short_help="delete templates and VMs",
)

app.add_typer(
    ezua.app,
    no_args_is_help=True,
    name="ezua",
    short_help="Manage Ezmeral Unified Analytics installation",
)

app.add_typer(
    ezdf.app,
    no_args_is_help=True,
    name="ezdf",
    short_help="Manage Ezmeral Data Fabric installation",
)


# ::GLOBALS --------------------------------------------------------------------- #
PKG_NAME = "ezlabctl"


# ::CORE LOGIC --------------------------------------------------------------------- #
# ::CLI ---------------------------------------------------------------------------- #


@app.command()
def info():
    """print usage"""
    print(
        """
        Manage lab VMs for Ezmeral products
        Select <target> parameters: 'pve' | 'vmw'
        Provide credentials: <host> <username> <password>
        <config> get/set: manage preferences and network/storage,
        <create> vm/template: deploy VMs based on template,
        <delete> vm/template: delete selected VMs and templates,
        <install> df/ua: install Ezmeral product,
        <prepare>: prepare VM(s) for Ezmeral product,
        <info>: print this message
        """
    )


@app.command(
    no_args_is_help=True,
    name="prepare",
    short_help="Run pre-requisites on VM for specified product",
)
def prepare(
    vm_ip: Annotated[str, typer.Option("--ip", "-i")],
    vm_name: Annotated[str, typer.Option("--name", "-n", help="short name for the host")],
    product: Annotated[
        str,
        typer.Option(
            "--product",
            "-p",
        ),
    ],
):
    """
    Configure vm for given product

    <ip>        : IP address
    <name>      : Short hostname
    <product>   : ezdf | ezua | generic
    """

    if prepare_vm(
        vm_name=vm_name,
        vm_ip=vm_ip,
        product_code=product,
    ):
        print(f"{vm_name} configured")


def _ensure_configured():
    # ensure the system is configured
    if not os.path.isfile(ini_file):
        print("Configuration is required.")
        config.set()
    if len(config.get().keys()) == 0:
        print("Not configured! run `ezlabctl config set` to start...")
        exit(-1)


def _version_callback(value: bool):
    if value:
        from . import __version__

        typer.echo(f"ezlabctl: v{__version__}")
        raise typer.Exit()


# ::EXECUTE ------------------------------------------------------------------------ #
@app.callback()
def run(
    version: Optional[bool] = typer.Option(None, "--version", "-v", callback=_version_callback, is_eager=True),
) -> None:
    return


def enter():
    _ensure_configured()
    app()



import configparser
from typing import Annotated
from questionary import password, text, path as qpath
from rich import print

import typer

from ezlabctl.parameters import ini_file

app = typer.Typer(add_completion=True, no_args_is_help=True)

config = configparser.ConfigParser()

config_keys = [
    {
        "title": "VM Network CIDR",
        "section": "NETWORK",
        "key": "vm_cidr",
        "default": "192.168.0.0/24",
        "type": text,
    },
    {
        "title": "VM Network Gateway",
        "section": "NETWORK",
        "key": "vm_gateway",
        "default": "192.168.0.1",
        "type": text,
    },
    {
        "title": "VM DNS Server",
        "section": "NETWORK",
        "key": "vm_nameserver",
        "default": "192.168.0.1",
        "type": text,
    },
    {
        "title": "VM DNS Domain",
        "section": "NETWORK",
        "key": "vm_domain",
        "default": "ez.lab",
        "type": text,
    },
    # {
    #     "title": "DNS API",
    #     "section": "NETWORK",
    #     "key": "dns_api",
    #     "default": "",
    #     "type": text,
    # },
    {
        "title": "PROXY",
        "section": "NETWORK",
        "key": "proxy",
        "default": "",
        "type": text,
    },
    {
        "title": "YUM_PROXY",
        "section": "NETWORK",
        "key": "yum_proxy",
        "default": "",
        "type": text,
    },
    {
        "title": "MAPR_REPO",
        "section": "NETWORK",
        "key": "mapr_repo",
        "default": "",
        "type": text,
    },
    {
        "title": "VM admin username",
        "section": "CLOUDINIT",
        "key": "vm_username",
        "default": "ezmeral",
        "type": text,
    },
    {
        "title": "VM admin password",
        "section": "CLOUDINIT",
        "key": "vm_password",
        "default": "",
        "type": password,
    },
    {
        "title": "SSH Public Key path",
        "section": "CLOUDINIT",
        "key": "ssh_keyfile",
        "default": "~/.ssh/id_rsa.pub",
        "type": qpath,
    },
    {
        "title": "Rocky Base Image",
        "section": "CLOUDIMG",
        "key": "rocky",
        "default": "Rocky-8-GenericCloud.latest.x86_64.qcow2",
        "type": text,
    },
    {
        "title": "RHEL8 Base Image",
        "section": "CLOUDIMG",
        "key": "rhel",
        "default": "rhel-8.8-cloudvm-qemu.qcow2",
        "type": text,
    },
]


def key_default(key):
    """
    Returns the existing or default value for the key
    """
    return config[key["section"]][key["key"]] if config.has_option(key["section"], key["key"]) else key["default"]


@app.command()
def set():
    config.read(ini_file)
    print(f"using {ini_file}")

    # pre-populate config if values are missing
    if "NETWORK" not in config.sections():
        config.add_section("NETWORK")
    if "CLOUDINIT" not in config.sections():
        config.add_section("CLOUDINIT")
    if "CLOUDIMG" not in config.sections():
        config.add_section("CLOUDIMG")

    try:
        for key in config_keys:
            config[key["section"]][key["key"]] = key["type"](
                key["title"],
                default=key_default(key),
            ).ask(
                kbi_msg=f"Reverting to default {key_default(key)}"
            ) or key_default(key)
            # add trailing slash for mapr_repo
            if key["key"] == "mapr_repo":
                if not config[key["section"]]["mapr_repo"].endswith("/"):
                    config[key["section"]]["mapr_repo"] += "/"
            # remove trailing slash from yum_repo
            elif key["key"] == "yum_proxy":
                if config[key["section"]]["yum_proxy"].endswith("/"):
                    config[key["section"]]["yum_proxy"] = config[key["section"]]["yum_proxy"].rstrip("/")

    except TypeError as error:
        print(error)
        print("Incorrect entry for settings")
        exit(1)

    with open(ini_file, "w") as configfile:
        config.write(configfile)


@app.command()
def get(out: Annotated[bool, typer.Option("--print", "-p", help="Print the config")] = False):
    config.read(ini_file)
    dictionary = {s: dict(config.items(s)) for s in config.sections()}
    # check if all settings are provided

    if out:
        print(dictionary)
    return dictionary



import os

from ezlab.utils import find_app

ezfab_yaml = os.path.abspath("ezfab.yaml")
orchestratorkubeconfig = os.path.abspath("ezfab-orchestrator-kubeconfig")
workloadkubeconfig = os.path.abspath("ezfab-workload-kubeconfig")
workloadcr = os.path.abspath("ezkf-workload-deploy-cr.yaml")
ezfabricctl = find_app("ezfabricctl")
ezfabrelease = os.path.abspath("ezfab-release.tgz")


ez_cluster_name = "ezlab"


# def init_orchestrator(
#     ip: str,
#     username: str,
#     privatekey: str,
#     ezfabrelease: str = ezfabrelease,
# ):
#     """Install UA orchestrator

#     Args:
#         ip (str): orchestrator IP address
#         ezfabrelease (str): full path to ezfab-release.tgz file
#     """

#     # write_ezfabyaml(username=username, privatekey=privatekey, filename=ezfab_yaml, purpose="coordinator_init", orchestrator_ip=ip)
#     ezfab_yaml = importlib_resources.files("ezlab").joinpath("ezkf-input.yaml")

#     if ezfabricctl is None:
#         return f"ezfabricctl not found, check if you get ezfabricctl_{platform.system().lower()}_amd64 binary from the installer container and marked it as executable."

#     if is_file(ezfabrelease):
#         yield "Starting coordinator deployment..."
#         # ezfabricctl orchestrator init --releasepkg /tmp/ezfab-release.tgz --input /tmp/ezkf-input.yaml --status /tmp/ezkf-orchestrator/status.txt --save-kubeconfig /tmp/ezkf-orchestrator/mgmt-kubeconfig
#         for out in execute(f"{ezfabricctl} o init -p {ezfabrelease} -i {ezfab_yaml} --save-kubeconfig {orchestratorkubeconfig}"):
#             yield out
#     else:
#         return "Provide full path to ezfab-release.tgz file"


# def add_hosts_to_pool(
#     worker_ips: list[str],
#     kubeconf: str = orchestratorkubeconfig,
# ):
#     """Add hosts to UA pool

#     Args:
#         worker_ips (List(str)): hostname/IP address to add (can be provided multiple times)
#         kubeconf (str): path to orchestrator kubeconfig
#     """

#     if ezfabricctl is None:
#         return f"ezfabricctl not found, check if you get ezfabricctl_{platform.system().lower()}_amd64 binary from the installer container and marked it as executable."

#     write_ezfabyaml(filename=ezfab_yaml, purpose="poolhost_init", workers=worker_ips)

#     # ezfabricctl poolhost init --input /tmp/hostPoolConfig.yaml --orchestrator-kubeconfig /tmp/ezkf-orchestrator/mgmt-kubeconfig --status /tmp/workload/hostPoolConfigStatus.txt
#     for out in execute(f"{ezfabricctl} ph init -i {ezfab_yaml} -c {os.path.abspath(kubeconf)}"):
#         print(out)


# def create_workload_cluster(
#     orchestrator_ip: str,
#     kubeconf: str = orchestratorkubeconfig,
#     cluster_name: str = ez_cluster_name,
# ):
#     """Create UA workload cluster

#     Args:
#         orchestrator (str): controller IP address
#         kubeconf (str): path to orchestrator kubeconfig
#     """

#     if ezfabricctl is None:
#         return f"ezfabricctl not found, check if you get ezfabricctl_{platform.system().lower()}_amd64 binary from the installer container and marked it as executable."

#     # use full path for the config file
#     kubeconf = os.path.abspath(kubeconf)

#     write_ezfabyaml(
#         filename=ezfab_yaml,
#         purpose="workload_init",
#         orchestrator_ip=orchestrator_ip,
#     )

#     # ezfabricctl workload init --input /tmp/clusterConfig.yaml --orchestrator-kubeconfig /tmp/ezkf-orchestrator/mgmt-kubeconfig --status /tmp/workload/clusterConfigStatus.txt
#     for out in execute(f"{ezfabricctl} workload init -i {ezfab_yaml} -c {kubeconf}"):
#         print(out)

#     # If the orchestrator is ready
#     if is_file(kubeconf):
#         # Save the workload kubeconfig
#         for out in execute(f"{ezfabricctl} workload get kubeconfig -n {cluster_name} -c {kubeconf} --save-kubeconfig {workloadkubeconfig}"):
#             print(out)
#         print(f"Workload kubeconfig saved as {workloadkubeconfig}")
#     else:
#         return f"Orchestrator kubeconfig file {kubeconf} not found!"


# def applyCr(
#     domain: str,
#     username: str,
#     password: str,
#     cluster_name: str = ez_cluster_name,
# ):
#     # Deploy the workload CR
#     k8s_config.load_kube_config(config_file=orchestratorkubeconfig)

#     # with open(workloadkubeconfig, "r") as f:
#     #     kubeconfig = toB64(f.read())

#     # k8s_config = {
#     #     "apiVersion": "v1",
#     #     "kind": "Secret",
#     #     "metadata": {
#     #         "name": "k8sconfig",
#     #         "namespace": cluster_name
#     #     },
#     #     "data": {
#     #         "kubeconfig": kubeconfig
#     #     }
#     # }

#     v1 = k8s_client.CoreV1Api()
#     secrets = v1.list_namespaced_secret(cluster_name)
#     if "authconfig" not in [secret.metadata.name for secret in secrets.items]:
#         admin = {
#             "admin_user": {
#                 "username": username,
#                 "fullname": "Ezlab Admin User",
#                 "email": f"{username}@{domain}",
#                 "password": password,
#             }
#         }
#         auth_config = {
#             "apiVersion": "v1",
#             "kind": "Secret",
#             "metadata": {"name": "authconfig", "namespace": cluster_name},
#             "data": {"internal_auth": toB64(str(admin))},
#         }
#         v1.create_namespaced_secret(cluster_name, body=auth_config)

#     api = k8s_client.CustomObjectsApi()

#     if cluster_name not in [
#         deployment["metadata"]["name"]
#         for deployment in getDeploymentCRs(api, cluster_name)
#     ]:
#         cr = api.create_namespaced_custom_object(
#             namespace=cluster_name,
#             group="ezkfops.hpe.ezkf-ops.com",
#             version="v1alpha1",
#             plural="ezkfworkloaddeploys",
#             body=get_ezkfcr_json(cluster_name=cluster_name),
#         )
#         if cluster_name in cr["metadata"]["name"]:
#             print(f"Deployment created for {cluster_name}")
#     else:
#         print(f"{cluster_name} exists, remove from EzkfWorkloadDeploy CR to retry")

#     return "This process doesn't work with the error: ERROR: in cluster_deploy - Error creating serializer: {'deployname': [ErrorDetail(string='This field may not be null.', code='null')], 'domainname': [ErrorDetail(string='This field may not be null.', code='null')], 'deployallapps': [ErrorDetail(string='This field may not be null.', code='null')], 'deployallinfra': [ErrorDeta..."


# def remove(
#     host: str,
#     ezfabrelease: str = ezfabrelease,
#     kubeconf: str = orchestratorkubeconfig,
# ):
#     """Remove UA

#     Args:
#         host (str): hostname/IP address to destroy
#     """

#     if ezfabricctl is None:
#         return f"ezfabricctl not found, check if you get ezfabricctl_{platform.system().lower()}_amd64 binary from the installer container and marked it as executable."

#     write_ezfabyaml(filename=ezfab_yaml, purpose="coordinator_init", orchestrator_ip=host)

#     if is_file(ezfabrelease):
#         print("Destroying EZUA...")
#         for out in execute(f"{ezfabricctl} o destroy --force -p {ezfabrelease} -i {ezfab_yaml} -c {os.path.abspath(kubeconf)}"):
#             print(out)
#         print("Removing kubeconfig and cr files")
#         for file in [orchestratorkubeconfig, workloadkubeconfig, ezfab_yaml]:
#             if is_file(file):
#                 os.remove(file)
#             else:
#                 print(f"{os.path.basename(file)} not found, skipping...")

#     else:
#         return "Provide full path to ezfab-release.tgz file"


# def write_ezfabyaml(
#     username: str = "",
#     privatekey: str = "",
#     filename: str = "",
#     purpose: str = None,
#     orchestrator_ip: str = "",
#     workers: list = [],
#     cluster_name: str = ez_cluster_name,
# ) -> bool:
#     """Generate yaml as input for ezfabctl
#     TODO: should be using yaml module for proper yaml generation
#     """

#     response = False

#     try:
#         # Just in case hostname/fqdn is provided for orchestrator (we assume IP address for the rest)
#         if not validate_ip(orchestrator_ip):
#             try:
#                 orchestrator_ip = gethostbyname(orchestrator_ip)
#             except Exception as error:
#                 return f"Cannot get host IP: {error}"

#         # get private key from public key path, assuming private key file has the same name with public key without the .pub suffix
#         # ssh_prvkey = ""
#         # with open(os.path.abspath(os.path.expanduser(pkeyfile)), "r") as f:
#         #     ssh_prvkey = f.read()

#         obj = {
#             "defaultHostCredentials": {
#                 "sshUserName": username,
#                 "sshPrivateKey": toB64(privatekey),
#             },
#             "airgap": {
#                 "registryUrl": app.storage.general[UA]["airgap_registry"],
#                 # "registryInsecure": True,
#             },
#             "hosts": [],
#         }

#         if purpose == "coordinator_init":
#             obj["hosts"] = [{"host": orchestrator_ip}]

#         elif purpose == "poolhost_init":
#             for host in workers:
#                 obj["hosts"].append(
#                     {
#                         "host": host,
#                         "labels": {
#                             "role": (
#                                 "worker"
#                                 if workers.index(host) > 0
#                                 # first one is the control plane (K8s master)
#                                 else "controlplane"
#                             )
#                         },
#                     }
#                 )

#         elif purpose == "workload_init":
#             obj["workload"] = {
#                 "deployEnv": "ezkube",
#                 "workloadType": "ezua",
#                 "clusterName": cluster_name,
#                 "resources": {"vcpu": 96},
#                 "controlplane": {"controlPlaneEndpoint": workers[0]},
#                 "controlPlaneHostLabels": {"role": "controlplane"},
#                 "workerHostLabels": {"role": "worker"},
#             }

#         with open(filename, "w") as f:
#             f.write(yaml.safe_dump(obj))

#     except Exception as error:
#         print(error)
#         response = False

#     return response


# def getDeploymentCRs(
#     api: k8s_client.CustomObjectsApi, namespace: str = ez_cluster_name
# ):
#     crs = api.list_namespaced_custom_object(
#         namespace=namespace,
#         group="ezkfops.hpe.ezkf-ops.com",
#         version="v1alpha1",
#         plural="ezkfworkloaddeploys",
#     )
#     if crs is not None and len(crs) > 0:
#         _, deployments, _, _ = crs.values()
#     else:
#         deployments = []

#     return deployments


# def get_ezkfcr_json(cluster_name: str = ez_cluster_name):
#     """Generate yaml for EzkfWorkloadDeploy CR"""

#     cr = {
#         "apiVersion": "ezkfops.hpe.ezkf-ops.com/v1alpha1",
#         "kind": "EzkfWorkloadDeploy",
#         "metadata": {"name": cluster_name, "namespace": cluster_name},
#         "spec": {
#             "clustername": cluster_name,
#             "deployallapps": True,
#             "deployallinfra": True,
#             "deployenv": "ezkube",
#             "deploytarget": "pph",
#             "domainname": "testdom.com",
#             "workloadtype": "ezua",
#             "authconfig": {"secret_name": "authconfig"},
#         },
#     }
#     # "k8sconfig": { "secret_name": "k8sconfig" },
#     # "tlsconfig":{ "secret_name": "tlsconfig" },
#     # TODO: Not sure if this is needed or works
#     # if vm_proxy != "":
#     #     cr["spec"]["proxy"] = {
#     #         "httpProxy": vm_proxy,
#     #         "httpsProxy": vm_proxy,
#     #         "noProxy": get_proxy_environment().split("no_proxy=")[1],
#     #     }

#     # # TODO: Not tested
#     # if "airgap_registry" in defaults["NETWORK"]:
#     #     cr["spec"]["airgap"] = {
#     #         "secret_name": "airgap",
#     #         "registryUrl": {defaults["NETWORK"]["airgap_registry"]},
#     #         "registryInsecure": {
#     #             "true"
#     #             if defaults["NETWORK"]["airgap_registry"].split("://")[0] == "http"
#     #             else "false"
#     #         },
#     #     }

#     return cr



#!/usr/bin/env python3
"""
DESCRIPTION:
    Deploy VMs on my lab and configure them for Ezmeral
USAGE EXAMPLE:
    > ezlab for GUI, ezlabctl for command line
"""

import logging
from nicegui import Client, ui, app, native, binding
import requests

import ezlab
from ezlab.parameters import DF, SUPPORTED_HVES, UA
from ezlab.pages import ezmeral
from ezlab.pages import settings
from ezlab.pages import targets
from ezlab.utils import README, LogElementHandler, status_container, toggle_log


logger = logging.getLogger()
ch = logging.StreamHandler()
ch.setFormatter(
    logging.Formatter(
        "%(asctime)s:%(levelname)s:%(name)s:%(funcName)s:%(lineno)d: %(message)s",
        datefmt="%H:%M:%S",
    )
)
logger.addHandler(ch)
logger.setLevel(logging.DEBUG)

# https://sam.hooke.me/note/2023/10/nicegui-binding-propagation-warning/
binding.MAX_PROPAGATION_TIME = 0.05

# Initialize
# Top-level keys in config
for key in ["target", "ui", "config", UA, DF, *SUPPORTED_HVES]:
    if key not in app.storage.general.keys():
        app.storage.general[key] = {}

# Set initial connection
app.storage.general["target"]["connected"] = False

# DEBUG for passed args
app.native.start_args["debug"] = True


def graceful_fail(exc: Exception):
    print("gracefully dying...")
    logger.debug("Exception: %s", exc)
    app.storage.user["busy"] = False


app.on_exception(graceful_fail)


# Main page
@ui.page("/")
async def home(client: Client):
    # initial status
    app.storage.user["busy"] = False
    # and log window
    app.storage.user["showlog"] = False

    # Header
    with ui.header(elevated=True).classes("items-center justify-between") as header:
        # home button
        ui.label("Ezlab").classes("text-bold text-lg")

        ui.space()

        with ui.row().classes("items-center"):
            ui.label("Settings")
            ui.button(icon="download", on_click=settings.config_show().open)
            ui.button(icon="upload", on_click=settings.config_load().open)

    # Footer
    with ui.footer() as footer:
        with ui.row().classes("w-full place-items-center"):
            ui.button(icon="menu", on_click=toggle_log).props("flat text-color=white")
            ui.label("Log")
            ui.space()
            status_container()
            ui.space()
            ui.spinner("ios", size="2em", color="red").bind_visibility_from(app.storage.user, "busy")
            ui.icon("check_circle", size="2em", color="green").bind_visibility_from(app.storage.user, "busy", lambda x: not x)
        log = ui.log().classes("w-full h-48 bg-neutral-300/30").style("white-space: pre-wrap").bind_visibility(app.storage.user, "showlog")
        logger.addHandler(LogElementHandler(log, logging.INFO))

    # Content

    # README section
    with (
        ui.expansion(
            "Readme",
            icon="description",
            caption="how to use this app?",
        )
        .classes("w-full")
        .classes("text-bold")
    ):
        # readme = ""
        # with open(os.path.dirname(__file__) + "/USAGE.md", "r") as readmefile:
        #     readme = readmefile.read()

        ui.markdown(README)

    # Settings section
    settings.menu()

    ui.separator()

    # Infrastructure section
    targets.menu()

    ui.separator()

    # Ezmeral Products section
    ezmeral.menu()

    ui.separator()

    # cleanup the download route after the client disconnected
    # await client.disconnected()
    # app.routes[:] = [route for route in app.routes if route.path != CONFIG_DL_PATH]


# INSECURE REQUESTS ARE OK in LAB
requests.packages.urllib3.disable_warnings()
urllib_logger = logging.getLogger("urllib3.connectionpool")
urllib_logger.setLevel(logging.WARNING)

requests_log = logging.getLogger("requests.packages.urllib3")
requests_log.setLevel(logging.WARNING)
watcher_logger = logging.getLogger("watchfiles.main")
watcher_logger.setLevel(logging.FATAL)

paramiko_log = logging.getLogger("paramiko.transport")
paramiko_log.setLevel(logging.WARNING)

logger.debug("Running version: %s", ezlab.__version__)


# Entry point for the module
def enter():
    # ui.run(
    #     ezmeral_icon = """
    #     <svg width="48" height="24" viewBox="0 0 48 24" fill="none" xmlns="http://www.w3.org/2000/svg">
    #         <path d="M7 8H41V16H7V8Z" fill="#01A982"/>
    #         <path d="M1 8H7V16H1V8Z" fill="#00775B"/>
    #         <path d="M41 8H47V16H41V8Z" fill="#00775B"/>
    #         <path d="M7 16H41V22H7V16Z" fill="#00775B"/>
    #         <path d="M7 2H41V8H7V2Z" fill="#00C781"/>
    #         <path d="M1 8L7 2V8H1Z" fill="#01A982"/>
    #         <path d="M1 16L7 22V16H1Z" fill="#01A982"/>
    #         <path d="M47 8L41 2V8H47Z" fill="#01A982"/>
    #         <path d="M47 16L41 22V16H47Z" fill="#01A982"/>
    #     </svg>
    #     """
    #     title="Ezlab",
    #     dark=None,
    #     favicon=ezmeral_icon,
    #     storage_secret = ("ezmeralr0cks",)
    #     show=True,
    #     # reload=False,
    # )

    ui.run(
        title="Ezlab",
        dark=None,
        native=True,
        window_size=(1024, 1024),
        # frameless=True,
        storage_secret="ezmeralr0cks",
        reload=False,
        port=native.find_open_port(),
    )


# For development and debugs
if __name__ in {"__main__", "__mp_main__"}:
    print("EZLAB RUNNING IN DEV MODE")
    ui.run(
        title="Ezlab",
        dark=None,
        native=True,
        window_size=(1024, 1024),
        # frameless=True,
        storage_secret="ezmeralr0cks",
        reload=True,
        port=native.find_open_port(),
    )



import os
import cryptography
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa

from ezlab.parameters import SSHKEYNAME


def save_sshkey(key: str):
    # verify format and get private key
    private_key = get_privatekey_from_string(key)
    if private_key is None:
        return (False, "Not a known key format")

    private_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.OpenSSH,
        encryption_algorithm=serialization.NoEncryption(),
    )
    with open(SSHKEYNAME, "wb") as private_keyfile:
        private_keyfile.write(private_pem)
    os.chmod(SSHKEYNAME, 0o600)

    public_pem = (
        private_key.public_key()
        .public_bytes(
            encoding=serialization.Encoding.OpenSSH,
            format=serialization.PublicFormat.OpenSSH,
        )
        .decode("utf-8")
    )

    with open(f"{SSHKEYNAME}.pub", "w") as public_keyfile:
        public_keyfile.write(public_pem)

    return (True, private_pem.decode())


def get_privatekey_from_string(key: str):
    try:
        # try OpenSSL format (RSA?)
        return serialization.load_pem_private_key(key.encode(), password=None)
    except ValueError:
        # try OpenSSH format
        try:
            return serialization.load_ssh_private_key(key.encode(), password=None)
        except (ValueError, TypeError, cryptography.exceptions.UnsupportedAlgorithm):
            return None
    except Exception as error:
        print(error)
        return None


def get_ssh_pubkey_from_privatekey(privatekey):
    try:
        return (
            privatekey.public_key()
            .public_bytes(
                encoding=serialization.Encoding.OpenSSH,
                format=serialization.PublicFormat.OpenSSH,
            )
            .decode()
        )
    except Exception as error:
        print(error)
        return None


def get_rsa_private_key(privatekey: serialization.SSHPrivateKeyTypes):
    if privatekey is not None:
        return privatekey.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption(),
        )
    else:
        return None


async def create_sshkey():
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
    private_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.OpenSSH,
        encryption_algorithm=serialization.NoEncryption(),
    ).decode("utf-8")
    await save_sshkey(private_pem)



# App parameters

TASK_FINISHED = "EZTASK_FINISHED"  # used as sentinel for queues

VMWARE = "VMware"
PVE = "Proxmox"
AWS = "AWS"
AZURE = "Azure"
KVM = "KVM"

SUPPORTED_HVES = [
    PVE,
    # VMWARE,
    KVM,
    # AWS
    # AZURE
]

UA = "EZUA"
DF = "EZDF"
DFAAS = "DFaaS"
GENERIC = "Generic"
DFCLIENT = "Client"

MAPR_CORE_VERSION = "7.6.1"
MAPR_MEP_VERSION = "9.2.1"

SSHKEYNAME = "ezlab-key"

EZNODES = [
    {
        "name": "ua-control",
        "product": UA,
        "cores": 8,
        "memGB": 32,
        "os_disk_size": 400,
        "data_disk_size": 500,
        "swap_disk_size": 8,
        "no_of_disks": 1,
        "count": 2,
    },
    {
        "name": "ua-worker",
        "product": UA,
        "cores": 32,
        "memGB": 128,
        "os_disk_size": 300,
        "data_disk_size": 500,
        "swap_disk_size": 32,
        "no_of_disks": 2,
        "count": 3,
    },
    {
        "name": "df-singlenode",
        "product": DF,
        "cores": 8,
        "memGB": 64,
        "os_disk_size": 240,
        "data_disk_size": 200,
        "swap_disk_size": 16,
        "no_of_disks": 1,
        "count": 1,
    },
    {
        "name": "df-5nodes",
        "product": DF,
        "cores": 8,
        "memGB": 32,
        "os_disk_size": 240,
        "data_disk_size": 200,
        "swap_disk_size": 8,
        "no_of_disks": 1,
        "count": 5,
    },
    {
        "name": "dfaas-3nodes",
        "product": DFAAS,
        "cores": 16,
        "memGB": 32,
        "os_disk_size": 240,
        "data_disk_size": 200,
        "swap_disk_size": 8,
        "no_of_disks": 1,
        "count": 3,
    },
    {
        "name": "generic",
        "product": GENERIC,
        "cores": 1,
        "memGB": 2,
        "os_disk_size": 30,
        "swap_disk_size": 2,
        "no_of_disks": 0,
        "count": 1,
    },
]

SETUP_COMMANDS = {
    DFAAS: [
        # "sudo useradd -md /home/mapr -u 5000 -U -s /bin/bash mapr",
        # "echo 'mapr ALL=(ALL) NOPASSWD:ALL' | sudo tee /etc/sudoers.d/mapr; sudo chown root:root /etc/sudoers.d/mapr; sudo chmod 440 /etc/sudoers.d/mapr",
        # "echo mapr:mapr | sudo chpasswd",
    ],
    DF: [
        # "sudo chown root:root /etc/sudoers.d/mapr -- 2>/dev/null && sudo chmod 440 /etc/sudoers.d/mapr",
        "sudo sed -i 's/^SELINUX=.*/SELINUX=disabled/' /etc/selinux/config",
        "sudo sysctl vm.swappiness=1 >/dev/null",
        "echo 'vm.swappiness=1' | sudo tee /etc/sysctl.d/mapr.conf >/dev/null",
        "echo 'umask 0022' | sudo tee /etc/profile.d/mapr.sh >/dev/null",
        "echo 'LD_LIBRARY_PATH=$LD_LIBRARY_PATH:/opt/mapr/lib' | sudo tee -a /etc/profile.d/mapr.sh >/dev/null",
        "sudo systemctl disable --now numad",
        "sudo dnf install -y -q wget jq",
    ],
    UA: [
        # "sudo dnf update -q -y",
        # "sudo subscription-manager repos --enable rhel-8-for-x86_64-highavailability-rpms",
        "sudo dnf -q -y install nfs-utils policycoreutils-python-utils conntrack-tools jq tar >/dev/null",
        "sudo dnf --setopt=tsflags=noscripts install -y -q iscsi-initiator-utils >/dev/null",
        'echo "InitiatorName=$(/sbin/iscsi-iname)" | sudo tee -a /etc/iscsi/initiatorname.iscsi >/dev/null',
        "sudo systemctl enable --now iscsid 2>&1",
        # "sudo systemctl disable --now firewalld 2>&1",
        "sudo modprobe ip_tables",
        "sudo sed -i 's/FirewallBackend=.*/FirewallBackend=iptables/' /etc/firewalld/firewalld.conf && sudo systemctl restart firewalld",
        "sudo ethtool -K eth0 tx-checksum-ip-generic off >/dev/null",
    ],
    GENERIC: [
        # "sudo dnf remove -y -q cloud-init",
        # disable subscription manager - TODO: needs testing
        "[ -f /etc/yum/pluginconf.d/product-id.conf ] && sudo sed -i 's/^enabled=1/enabled=0/' /etc/yum/pluginconf.d/product-id.conf",
        "[ -f /etc/yum/pluginconf.d/subscription-manager.conf ] && sudo sed -i 's/^enabled=1/enabled=0/' /etc/yum/pluginconf.d/subscription-manager.conf",
        # since PVE cloudinit doesn't support extra args, ensure this is configured
        "sudo sed -i 's/^[^#]*PasswordAuthentication[[:space:]]no/PasswordAuthentication yes/' /etc/ssh/sshd_config",
        "sudo systemctl restart sshd",
        # disable ipv6 returns for getent hosts
        "sudo sed -i 's/myhostname//g' /etc/nsswitch.conf",
    ],
}

INSTALL_COMMANDS = {
    DF: [
        """set -euo pipefail
        [ -f /opt/mapr/installer/bin/mapr-installer-cli ] && echo 'Starting installation, this may take ~30 minutes...'
        echo y | sudo /opt/mapr/installer/bin/mapr-installer-cli install -nvp -t /tmp/mapr.stanza -u mapr:mapr@127.0.0.1:9443
        """
    ],
    UA: [
        "ezfabricctl orchestrator init --releasepkg /tmp/ezfab-release.tgz --input /tmp/ezkf-input.yaml --status /tmp/ezkf-orchestrator/status.txt --save-kubeconfig /tmp/ezkf-orchestrator/mgmt-kubeconfig",
        "ezfabricctl poolhost init --input /tmp/hostPoolConfig.yaml --orchestrator-kubeconfig /tmp/ezkf-orchestrator/mgmt-kubeconfig --status /tmp/workload/hostPoolConfigStatus.txt",
        "ezfabricctl workload init --input /tmp/clusterConfig.yaml --orchestrator-kubeconfig /tmp/ezkf-orchestrator/mgmt-kubeconfig --status /tmp/workload/clusterConfigStatus.txt",
        "ezfabricctl workload get kubeconfig --input /tmp/clusterConfig.yaml --orchestrator-kubeconfig /tmp/ezkf-orchestrator/mgmt-kubeconfig --status /tmp/workload/clusterConfigStatus.txt",
        "kubectl --kubeconfig=./tmp/workload/cluster-kubeconfig apply -f ./tmp/ezkfWorkloadDeploy.yaml",
    ],
    DFCLIENT: [
        "sudo dnf update -y -q",
        "sudo useradd -md /home/mapr -u 5000 -U -s /bin/bash mapr",
        "echo mapr:mapr | sudo chpasswd",
        "sudo dnf install -y -q mapr-core mapr-client mapr-posix-client-basic java-11-openjdk",
    ],
}

NO_PROXY = "10.96.0.0/12,10.224.0.0/16,10.43.0.0/16,192.168.0.0/16,.external.hpe.local,localhost,.cluster.local,.svc,.default.svc,127.0.0.1,169.254.169.254,.{vm_domain},{vm_network},{no_proxy}"

DF_SECURE_FILES = [
    "/opt/mapr/conf/ssl_truststore",
    "/opt/mapr/conf/ssl_truststore.p12",
    "/opt/mapr/conf/ssl_truststore.pem",
    "/opt/mapr/conf/maprtrustcreds.conf",
    "/opt/mapr/conf/maprtrustcreds.jceks",
    "/opt/mapr/conf/ssl_keystore-signed.pem",
]



# Config show dialog
import json
import logging
from nicegui import app, ui
from ezlab.parameters import DF
from ezlab.pages import ezmeral

from ezlab.pages.sshkeys import ssh_keys_ui
from ezlab.pages import targets
from ezlab.pages import settings
from ezlab.utils import CONFIG_DL_PATH

logger = logging.getLogger("ezlab.ui.settings")


def save_config(val: str, dialog):
    try:
        for key, value in json.loads(val.replace("\n", "")).items():
            app.storage.general[key] = value
        for m in (ezmeral.menu, targets.menu, settings.menu):
            m.refresh()
        dialog.close()
        ui.notify("Settings loaded", type="positive")
    except (TypeError, json.decoder.JSONDecodeError, ValueError) as error:
        ui.notify("Not a valid json", type="negative")
        logger.warning(error)


def config_show():
    with ui.dialog() as config_show, ui.card().classes("w-full h-full"):
        config_json = json.dumps(app.storage.general, indent=2)
        ui.code(config_json, language="json").classes("w-full text-wrap")
        with ui.row():
            ui.button(
                "Save",
                icon="save",
                on_click=lambda: ui.download(CONFIG_DL_PATH),
            )
            ui.button(
                "Close",
                icon="cancel",
                on_click=config_show.close,
            )
    return config_show


# Config load dialog
def config_load():
    with ui.dialog() as config_load, ui.card().classes("w-full h-full"):
        jsontext = ui.textarea().props("stack-label=json autogrow filled").classes("h-dvh w-full text-wrap")
        with ui.row():
            ui.button(
                "Save",
                on_click=lambda _: save_config(
                    jsontext.value,
                    config_load,
                ),
                icon="save",
            )
            ui.button(
                "Close",
                icon="cancel",
                on_click=config_load.close,
            )

    return config_load


@ui.refreshable
def menu():
    with (
        ui.expansion(
            "Settings",
            icon="settings",
            caption="for your environment",
        )
        .classes("w-full")
        .classes("text-bold") as settings
    ):
        ui.checkbox("Dry Run", value=False).bind_value(app.storage.general["config"], "dryrun")
        ui.label("Network Settings").classes("text-bold")
        with ui.row():
            ui.input("VM Network", placeholder="10.1.1.0/24").bind_value(app.storage.general["config"], "cidr")
            ui.input("Gateway", placeholder="10.1.1.1").bind_value(app.storage.general["config"], "gateway")
            ui.input("Name Server", placeholder="10.1.1.1").bind_value(app.storage.general["config"], "nameserver")
            ui.input("Domain", placeholder="ez.lab").bind_value(app.storage.general["config"], "domain")
            ui.input("HTTP Proxy", placeholder="").bind_value(app.storage.general["config"], "proxy")

        ui.label("Repositories").classes("text-bold")
        with ui.row().classes("w-full"):
            # YUM Repo
            ui.input("YUM Repo", placeholder="").bind_value(app.storage.general["config"], "yumrepo")

            ui.input("EPEL Repo", placeholder="").bind_value(app.storage.general["config"], "epelrepo")
            # ui.input("Insecure Registry", placeholder="").bind_value(
            #     app.storage.general["config"], "airgap_registry"
            # )

        # MapR Repository Configuration
        with ui.row().classes("w-full"):
            ui.label("MapR Repo").classes("self-center")
            switch = ui.switch("Use local", value=False).props("left-label").bind_value(app.storage.general[DF], "maprrepoislocal")

            # Using default HPE repository
            with ui.row().bind_visibility_from(switch, "value", lambda x: not x).classes("w-full items-justify"):
                ui.input("HPE Passport e-mail").bind_value(app.storage.general[DF], "maprrepouser")
                ui.input("HPE Passport token", password=True, password_toggle_button=True).bind_value(app.storage.general[DF], "maprrepotoken")

            # Using local repository
            with ui.row().bind_visibility_from(switch, "value"):
                ui.input(
                    "MapR Repo",
                    placeholder="https://repo.ez.lab/mapr/",
                ).bind_value(app.storage.general[DF], "maprlocalrepo")
                authlocal = ui.switch("Authenticate", value=True).bind_value(app.storage.general[DF], "maprlocalrepoauth")
                with ui.row():
                    ui.input("Username", password=True).bind_value(app.storage.general[DF], "maprlocalrepousername").bind_visibility(
                        authlocal, "value"
                    )
                    ui.input("Password", password=True).bind_value(app.storage.general[DF], "maprlocalrepopassword").bind_visibility(
                        authlocal, "value"
                    )

        ui.label("Cloudinit Settings").classes("text-bold")
        with ui.row():
            ui.input("username", placeholder="ezmeral").bind_value(app.storage.general["config"], "username")
            ui.input("password", password=True, password_toggle_button=True).bind_value(app.storage.general["config"], "password")

        ui.label("SSH Key").classes("text-bold")
        ssh_keys_ui()

    settings.bind_value(app.storage.general["ui"], "settings")



import os

from ezlab.parameters import DF, GENERIC, UA

ini_file = os.path.expanduser("~/.ezconfig.ini")


ez_files = {
    DF: {
        "/etc/sysctl.d/mapr.conf": "vm.swappiness=1\n",
        "/etc/profile.d/mapr.sh": "umask 0022\n",
    },
    UA: {},
    GENERIC: {},
}


ez_installers = {
    DF: {
        "image": "maprtech/edf-seed-container:latest",
        "start_installer": "./datafabric_container_setup.sh",
        "url": "https://localhost:8443/app/dfui/#/installer",
        "commands": [],
    },
    UA: {
        "start_installer": "./start_ezua_installer_ui.sh",
        # "image": "marketplace.us1.greenlake-hpe.com/ezua/ezua/hpe-ezua-installer-ui:latest",
        "image": "us.gcr.io/mapr-252711/hpe-ezua-installer-ui:1.2.0-d16afb0",
        "url": "http://localhost:8080",
    },
    GENERIC: {},
}



import asyncio
from nicegui import app, ui
from proxmoxer import ProxmoxAPI
from libvirt import virConnect
import ezlab.utils as utils

from ezlab.parameters import KVM, PVE, SUPPORTED_HVES

from ezlab.pages.vms import new_vm_ui
from ezinfra import kvm, pve


@ui.refreshable
def menu():

    with ui.expansion("VMs", icon="computer", caption="create and configure").classes("w-full").classes("text-bold") as virtualmachines:
        with ui.row().classes("w-full items-center justify-between"):
            # Login dialog
            with ui.dialog() as dialog, ui.card():
                t = ui.radio(options=SUPPORTED_HVES).classes("w-full").props("inline").bind_value(app.storage.general["target"], "hve")

                h = ui.input("Host").classes("w-full").bind_value(app.storage.general["target"], "host")

                u = ui.input("Username").classes("w-full").bind_value(app.storage.general["target"], "username")

                p = (
                    ui.input(
                        "Password",
                        password=True,
                        password_toggle_button=True,
                    )
                    .classes("w-full")
                    .bind_value(app.storage.general["target"], "password")
                    .bind_enabled_from(  # disable password entry for kvm, only ssh key authorisation works (or user need to enter password at the cli)
                        app.storage.general["target"], "hve", backward=lambda x: x != KVM
                    )
                )

                ui.button(
                    "Login",
                    on_click=lambda: dialog.submit((t.value, h.value, u.value, p.value)),
                )

            # Connection view
            with ui.row():
                ui.button(icon="cloud_done", on_click=lambda: loginbuttonaction(dialog)).bind_visibility_from(
                    app.storage.general["target"],
                    "connected",
                )
                ui.button(icon="cloud_off", on_click=lambda: loginbuttonaction(dialog)).bind_visibility_from(
                    app.storage.general["target"],
                    "connected",
                    lambda x: not x,
                )

            # new vm button
            ui.label().bind_text_from(
                app.storage.general["target"],
                "hve",
                backward=lambda x: f"{x} connected",
            ).bind_visibility_from(app.storage.general["target"], "connected")
            ui.button("New VM", on_click=new_vm_ui).bind_visibility_from(app.storage.general["target"], "connected").bind_enabled_from(
                app.storage.user, "busy", backward=lambda x: not x
            )

    virtualmachines.bind_value(app.storage.general["ui"], "virtualmachines")


async def loginbuttonaction(loginform: ui.dialog):
    if app.storage.general["target"]["connected"]:
        # TODO: disconnect process/cleanup
        app.storage.general["target"]["connected"] = False

    else:
        result = await loginform

        # modify password field if target is KVM, so emptiness check below can proceed without password
        if result and result[0] == KVM:
            result[3] == "nopasswordsupportforkvm"

        # emptiness check
        if result and all(result):
            connection = await connect(result)
            if connection:
                utils.ezapp.connection = connection
                app.storage.general["target"]["connected"] = True

        else:
            ui.notify("Missing information", type="warning")


async def connect(params: set):
    target, host, username, password = params
    try:
        if target == PVE:
            result = await asyncio.to_thread(pve.connect, host, username, password)

        elif target == KVM:
            result = await asyncio.to_thread(kvm.connect, host, username)

        # elif target == VMWARE:
        #     return await asyncio.to_thread(vmw.connect, host, username, password)

        else:
            result = None

        if isinstance(result, ProxmoxAPI) or isinstance(result, virConnect):
            app.storage.general["target"]["connected"] = True
            return result
        else:
            app.storage.general["target"]["connected"] = False
            ui.notify(f"FAILED: {result}", type="negative")

    except Exception as error:
        ui.notify(f"CONNECT ERROR: {error}", type="negative")



import logging
from nicegui import app, ui, run
from ezlab.pages.df import get_doc
from ezlab.parameters import EZNODES, KVM, PVE
from ezinfra.vms import clone
from ezinfra import kvm, pve
from ezlab.utils import get_new_vm_name


logger = logging.getLogger("ezlab.ui.vms")


def pve_newvm_dialog():

    @ui.refreshable
    def bridge_selector():
        app.storage.user["busy"] = True
        selector = (
            ui.select(
                options=app.storage.user.get("bridges", []),
                label="Base Image",
            )
            .classes("w-full")
            .props("inline")
            .bind_value(app.storage.general[PVE], "bridge")
        )
        app.storage.user["busy"] = False
        return selector

    def update_template():
        app.storage.user["bridges"] = pve.bridges(app.storage.general[PVE].get("template", None))
        bridge_selector.refresh()

    app.storage.user["busy"] = True

    templates = [x for x in pve.vms() if x["template"] and x["type"] == "qemu"]
    # networks = pve.networks(ezapp.connection)
    storage = pve.pools()

    app.storage.user["busy"] = False

    # vm settings dialog
    with ui.dialog() as dialog, ui.card():
        ui.select(
            options={x["id"]: f"{x['name']} ( {x['node']} )" for x in templates},
            label="Template",
            on_change=update_template,
        ).classes(
            "w-full"
        ).props("inline").bind_value(app.storage.general[PVE], "template")

        ui.select(
            options={x["id"]: f"{x['storage']} ( {x['node']} )" for x in storage},
            # if x["node"] == app.storage.user["template"]["node"]
            label="Data Disk Pool",
        ).classes("w-full").props("inline").bind_value(app.storage.general[PVE], "storage")

        # network_selector = (
        #     ui.select(
        #         options=[
        #             x["sdn"]
        #             for x in networks
        #             # if x["node"] == app.storage.user["template"]["node"]
        #         ],
        #         label="Network",
        #     )
        #     .classes("w-full")
        #     .props("inline")
        #     .bind_value(app.storage.general[PVE], "network")
        # )

        ui.select(
            options=[x["iface"] for x in app.storage.user.get("bridges", [])],
            label="Bridge",
        ).classes("w-full").props(
            "inline"
        ).bind_value(app.storage.general[PVE], "bridge")

        ui.select(
            options=[x["name"] for x in EZNODES],
            label="Node Type",
        ).classes("w-full").props(
            "inline"
        ).bind_value(app.storage.general[PVE], "eznode")

        ui.input(label="VM Name", placeholder="ezvm").classes("w-full").props("inline").bind_value(app.storage.general[PVE], "hostname")

        ui.input(
            label="First IP",
            placeholder=app.storage.general["config"]["gateway"],
        ).classes("w-full").props(
            "inline"
        ).bind_value(app.storage.general[PVE], "firstip")

        ui.button(
            "Create",
            on_click=lambda: dialog.submit(
                (
                    next(
                        (i for i in templates if i["id"] == app.storage.general[PVE]["template"]),
                        None,
                    ),
                    next(
                        (i for i in storage if i["id"] == app.storage.general[PVE]["storage"]),
                        None,
                    ),
                    app.storage.general[PVE]["bridge"],
                    next(
                        (i for i in EZNODES if i["name"] == app.storage.general[PVE]["eznode"]),
                        None,
                    ),
                    app.storage.general[PVE]["hostname"],
                    app.storage.general[PVE]["firstip"],
                )
            ),
        )

    return dialog


def kvm_newvm_dialog():
    @ui.refreshable
    def volume_selector():
        app.storage.user["busy"] = True
        selector = (
            ui.select(
                options=app.storage.user.get("volumes", []),
                label="Base Image",
            )
            .classes("w-full")
            .props("inline")
            .bind_value(app.storage.general[KVM], "baseimg")
        )
        app.storage.user["busy"] = False
        return selector

    def update_volumes():
        app.storage.user["volumes"] = kvm.volumes(app.storage.general[KVM].get("pool", "default"))
        volume_selector.refresh()

    # select dialog
    with ui.dialog() as dialog, ui.card():

        app.storage.user["busy"] = True

        pools = kvm.pools()
        bridges = kvm.bridges()

        app.storage.user["busy"] = False

        ui.select(
            options=[p.name() for p in pools],
            label="Pool",
            on_change=update_volumes,
        ).classes("w-full").props(
            "inline"
        ).bind_value(app.storage.general[KVM], "pool")

        # baseimg
        volume_selector()

        # volumes = kvm.volumes(app.storage.general[KVM].get("pool", "default"))
        # baseimg = (
        #     ui.select(
        #         options=volumes,
        #         label="Base Image",
        #     )
        #     .classes("w-full")
        #     .props("inline")
        #     .bind_value(app.storage.general[KVM], "baseimg")
        # )

        ui.select(
            options=bridges,
            label="Bridge",
        ).classes("w-full").props(
            "inline"
        ).bind_value(app.storage.general[KVM], "bridge")

        ui.select(
            options=[x["name"] for x in EZNODES],
            label="Node Type",
        ).classes("w-full").props(
            "inline"
        ).bind_value(app.storage.general[KVM], "eznode")

        ui.input(label="VM Name", placeholder="ezvm").classes("w-full").props("inline").bind_value(app.storage.general[KVM], "hostname")

        ui.input(
            label="First IP",
            placeholder="10.1.1.21",
        ).classes("w-full").props(
            "inline"
        ).bind_value(app.storage.general[KVM], "firstip")

        ui.button(
            "Create",
            on_click=lambda: dialog.submit(
                (
                    app.storage.general[KVM]["pool"],
                    app.storage.general[KVM]["baseimg"],
                    app.storage.general[KVM]["bridge"],
                    next(
                        (i for i in EZNODES if i["name"] == app.storage.general[KVM]["eznode"]),
                        None,
                    ),
                    app.storage.general[KVM]["hostname"],
                    app.storage.general[KVM]["firstip"],
                )
            ),
        )
    return dialog


async def new_vm_ui():

    if app.storage.general["target"]["hve"] == PVE:
        dialog = pve_newvm_dialog()
    elif app.storage.general["target"]["hve"] == KVM:
        dialog = kvm_newvm_dialog()
    else:
        ui.notify("not implemented")

    resources = await dialog

    if resources and all(resources):
        vm_count = resources[3]["count"]

        app.storage.user["busy"] = True
        dryrun = app.storage.general["config"].get("dryrun", True)

        result = await run.io_bound(
            clone,
            target=app.storage.general["target"]["hve"],
            resources=resources,
            settings=dict(app.storage.general["config"]),
            # vm_number=(0 if vmcount == 1 else count + 1),  # start from 1 not 0, use 0 for single vm
            vm_count=vm_count,
            dryrun=dryrun,
        )
        if dryrun:
            get_doc(result)
        elif result:
            logger.info("%s VMs ready: %s", vm_count, result)

        app.storage.user["busy"] = False

    else:
        ui.notify("VM creation cancelled", type="info")



from ezlab.parameters import NO_PROXY


def get_proxy_vars(proxy):
    if proxy[-1] != "/":
        proxy += "/"
    return {
        "HTTP_PROXY": proxy,
        "http_proxy": proxy,
        "HTTPS_PROXY": proxy,
        "https_proxy": proxy,
        "NO_PROXY": NO_PROXY,
        "no_proxy": NO_PROXY,
    }


def proxy_files(proxy_url, domain, cidr, noproxy):
    env = []
    prx = []
    for key, value in get_proxy_vars(proxy_url).items():
        val = value.format(vm_domain=domain, vm_network=cidr, no_proxy=noproxy)
        env.append(f"{key}={val}")
        prx.append(f"export {key}={val}")

    return [
        {"/etc/environment": "\n".join(env)},
        {"/etc/sysconfig/proxy": "\n".join(prx)},
    ]



import asyncio
import base64
from datetime import datetime, timedelta
from ipaddress import IPv4Address, ip_address
import logging
import os
import random
import shlex
import socket
import string
import subprocess
import importlib.resources
import io
import json
import time
import uuid

from fastapi.responses import StreamingResponse
from nicegui import app, ui


class ModuleContext:
    connection = None

    def __init__(self) -> None:
        pass


ezapp = ModuleContext()


class LogElementHandler(logging.Handler):
    """A logging handler that emits messages to a log element."""

    def __init__(self, element: ui.log, level: int = logging.NOTSET) -> None:
        self.element = element
        super().__init__(level)

    def emit(self, record: logging.LogRecord) -> None:
        # change log format for UI
        self.setFormatter(
            logging.Formatter(
                "%(asctime)s:%(levelname)s: %(message)s",
                datefmt="%H:%M:%S",
            )
        )
        try:
            msg = self.format(record)
            self.element.push(msg)
        except Exception:
            self.handleError(record)


# Read the usage
README = importlib.resources.read_text("ezlab", "USAGE.md")


# validation for non-zero entry
def nonzero(val):
    return len(val) != 0


# wrapper to make sync calls async-like
def fire_and_forget(f):
    def wrapped(*args, **kwargs):
        return asyncio.get_event_loop().run_in_executor(None, f, *args, *[v for v in kwargs.values()])

    return wrapped


# find app on system path
def find_app(name: str):
    """Return executable path for `name` on PATH or in CWD."""
    from shutil import which

    return which(cmd=name, path=f"{os.environ.get('PATH')}:{os.getcwd()}")


# execute app/script locally
def execute(cmd: str):
    proc = subprocess.Popen(
        args=shlex.split(cmd),
        stdout=subprocess.PIPE,
        universal_newlines=True,
        # env=sub_env,
    )

    while proc.stdout:
        output = proc.stdout.readline().strip()
        if output == "" and proc.poll() is not None:
            break
        if output:
            yield output
    rc = proc.poll()


# Base64 conversion for secrets etc
def toB64(string: str):
    return base64.b64encode(string.encode("ascii")).decode("utf-8")


# IP address validation
def validate_ip(ip_string: str) -> bool:
    try:
        return isinstance(ip_address(ip_string), IPv4Address)
    except:
        return False


# reliably get FQDN for given IP
def get_fqdn(ip_string: str) -> str:
    fqdn, _, _ = socket.gethostbyaddr(ip_string)
    return fqdn


# def get_host_ip(hostname: str):
#     info = socket.getaddrinfo(
#         hostname, 22, family=socket.AF_INET, proto=socket.IPPROTO_TCP
#     )
#     return info[0][4][0]


def is_file(file):
    return os.path.isfile(os.path.abspath(file))


def dt_from_iso(timestring: str):
    # Workaround since received timestring with " AM"/" PM" suffix is not parsed properly
    isPM = " PM" in timestring
    dt = datetime.strptime(timestring.replace(" AM", "").replace(" PM", ""), "%Y-%m-%dT%H:%M:%S.%f%z")
    return dt + timedelta(hours=12) if isPM else dt


def create_ci_iso(files: dict, destfile: str):
    from io import BytesIO
    import pycdlib

    metadata = files["meta-data"]
    userdata = files["user-data"]
    netconf = files["network-config"]

    iso = pycdlib.PyCdlib()
    iso.new(
        interchange_level=3,
        joliet=True,
        sys_ident="LINUX",
        rock_ridge="1.09",
        vol_ident="cidata",
    )

    metadata_file = BytesIO(metadata.encode())
    metadata_file.seek(0)
    iso.add_fp(
        metadata_file,
        len(metadata),
        "/METADATA.;1",
        rr_name="meta-data",
        joliet_path="/meta-data",
    )

    userdata_file = BytesIO(userdata.encode())
    iso.add_fp(
        userdata_file,
        len(userdata),
        "/USERDATA.;1",
        rr_name="user-data",
        joliet_path="/user-data",
    )

    netconf_file = BytesIO(netconf.encode())
    netconf_file.seek(0)
    iso.add_fp(
        netconf_file,
        len(netconf),
        "/NETWORK_.;1",
        rr_name="network-config",
        joliet_path="/network-config",
    )

    iso.write(destfile)
    iso.close()
    userdata_file.close()
    metadata_file.close()
    netconf_file.close()


# Define a download path for config downloads
CONFIG_DL_PATH = f"/config/{uuid.uuid4()}.json"


@app.get(CONFIG_DL_PATH)
def download(content: str = None):
    # by default downloading settings
    if content is None:
        content = app.storage.general

    string_io = io.StringIO(json.dumps(content))  # create a file-like object from the string

    headers = {"Content-Disposition": "attachment; filename=config.json"}
    return StreamingResponse(string_io, media_type="text/plain", headers=headers)


APP_STATUS_CONTAINER = []


@ui.refreshable
def status_container():
    with ui.row():
        for stat in APP_STATUS_CONTAINER:
            item, dest = stat
            ui.link(text=item, target=dest, new_tab=True).classes("text-emerald-400")
            ui.space()


def toggle_log():
    app.storage.user["showlog"] = not app.storage.user["showlog"]


def get_doc(content: dict):

    random_eof = "".join(random.choices(string.ascii_uppercase + string.digits, k=10))

    def indent(strings: str, spaces: int):
        return strings.strip().replace("\n", f"\n{' '*spaces}")

    settings = ""

    for key, value in content["settings"].items():
        settings += f"""
    #   {key}: {value if len(str(value)) < 80 else value[:77] + '...'}"""

    files = ""
    for filepath, filecontent in content["files"].items():
        files += f"""
    sudo cat <<{random_eof}>{filepath}
    {indent(filecontent,4)}
    {random_eof}
    """

    commands = ""
    for command in content["commands"]:
        commands += f"""
    {indent(command,4)}
    """

    doc = f"""
    #!/usr/bin/env bash

    # Ezlab Run {time.strftime('%Y%m%d%H%M%z')}

    # {content["task"]}

    # Settings
        {settings}

    # Write Files
        {files}

    # Run Commands
        {commands}

    # End of {content['task']}

"""

    with (
        ui.dialog().props("full-width") as dialog,
        ui.card().classes("w-full"),
    ):
        ui.code(doc, language="bash").classes("w-full text-wrap")
        with ui.row():
            ui.button(
                "Save",
                icon="save",
                on_click=lambda: download(doc),
            )
            ui.button(
                "Close",
                icon="cancel",
                on_click=dialog.close,
            )
    dialog.open()


def get_new_vm_name(name_pattern: str, count: int):
    # intelligently increase the hostname counter
    # if we have single host, no need to append a number
    # else, check if the hostname is already numbered, if so, add to that number
    # NOTE: it won't roll to tens (ie, if previous host was named vm39, next will be vm310, not vm40)
    lastchar = name_pattern[-1]
    if lastchar.isdigit():
        count = int(lastchar) + count
        newname = name_pattern[:-1]

    return newname + str(count)



# Generate cloud-init files


def cloudinit_userdata(
    hostname: str,
    fqdn: str,
    ssh_username: str,
    ssh_password: str,
    ssh_pubkey: str,
    swap_disk: str,
):
    return f"""
#cloud-config
hostname: {hostname}
manage_etc_hosts: true
fqdn: {fqdn}
user: {ssh_username}
password: {ssh_password}
ssh_authorized_keys:
  - {ssh_pubkey}
chpasswd:
  list: |
    {ssh_username}:{ssh_password}
    root:{ssh_password}
  expire: False
ssh_pwauth: true
disable_root: false

# users:
#   - default

# fs_setup:
#   - device: {swap_disk}
#     filesystem: swap
# mounts:
#   - ["{swap_disk}", "none", "swap", "sw,nofail,x-systemd.requires=cloud-init.service", "0", "0"]

"""


def cloudinit_metadata(hostname: str):
    return f"""
instance-id: {hostname}
local-hostname: {hostname}"""


def cloudinit_networkconfig(interface: str, ip_address: str, network_bits: int, gateway: str, nameserver: str, searchdomain: str):
    return f"""version: 1
config:
  - type: physical
    name: {interface}
    subnets:
      - type: static
        address: {ip_address}/{network_bits}
        gateway: {gateway}
        dns_nameservers:
          - {nameserver}
        dns_search:
         - {searchdomain}
"""



from ipaddress import ip_address
import os
import subprocess
from time import sleep
from urllib.parse import quote
import humanize
from questionary import confirm
from proxmoxer import ProxmoxAPI

from . import connect
from . import config

from .common import (
    ReturningThread,
    fail,
    get_ssh_key_file,
    multiselect_from_list,
    prepare_vm,
    select_from_list,
    start_tasks_delayed,
    wait_for_ssh,
    wait_tasks,
)


def pve_nextid(proxmox):
    return proxmox.cluster.nextid.get()


def get_connection(host, username, password):
    try:
        return ProxmoxAPI(
            host,
            user=username,
            password=password,
            verify_ssl=False,
        )

    except Exception as error:
        print(error)
        connect.fail("Connection error: " + str(type(error)))


def task_status(proxmox, node, task_id):
    return proxmox.nodes(node).tasks(task_id).status.get()


def template_name(t):
    return (" | ").join([t["name"], t["node"], t["type"]])


def vm_name(vm):
    return (" | ").join([vm["name"], vm["id"], vm["status"]])


def network_name(n):
    return (" | ").join([n["sdn"], n["node"], n["status"]])


def volume_name(v):
    return " | ".join(
        [
            v["storage"],
            v["node"],
            v["plugintype"],
            "/".join(
                [
                    humanize.naturalsize(v["disk"], binary=True),
                    humanize.naturalsize(v["maxdisk"], binary=True),
                ]
            ),
        ]
    )


def vms(proxmox):
    return proxmox.cluster.resources.get(type="vm")


def storage(proxmox):
    return proxmox.cluster.resources.get(type="storage")


def networks(proxmox):
    return proxmox.cluster.resources.get(type="sdn")


def vm_id(vm):
    return vm["id"]


def volume_id(volume):
    return volume["id"]


def network_id(network):
    return network["id"]


def vm_filter(fields):
    return fields["type"] == "qemu" and fields["template"] == 0


def template_filter(fields):
    return fields["type"] == "qemu" and fields["template"] == 1


def volume_filter(fields):
    return "images" in fields["content"].split(",")


def network_filter(fields):
    return True


def select_template(proxmox):
    return select_from_list(
        title="Template",
        list=[x for x in vms(proxmox) if template_filter(x)],
        name_generator=template_name,
        value_generator=vm_id,
    )


def select_storage(proxmox, host=None):
    return select_from_list(
        title="Data Storage",
        list=sorted(
            [x for x in storage(proxmox) if volume_filter(x) and x["node"] == host],
            key=lambda x: x["storage"],
        ),
        name_generator=volume_name,
        value_generator=volume_id,
    )


def select_network(proxmox, host=None):
    return select_from_list(
        title="Network",
        list=[x for x in networks(proxmox) if network_filter(x) and x["node"] == host],
        name_generator=network_name,
        value_generator=network_id,
    )


def select_network_bridge(proxmox, host):
    return select_from_list(
        title="Network Bridge",
        list=proxmox.nodes(host).network.get(type="any_bridge"),
        name_generator=lambda x: (" | ").join([x["iface"], x["cidr"]]),
        value_generator=lambda x: x["iface"],
    )


def select_vms(proxmox):
    return multiselect_from_list(
        title="Virtual Machines",
        list=[x for x in vms(proxmox) if vm_filter(x)],
        name_generator=vm_name,
        value_generator=vm_id,
    )


def select_templates(proxmox):
    return multiselect_from_list(
        title="Templates",
        list=[x for x in vms(proxmox) if template_filter(x)],
        name_generator=vm_name,
        value_generator=vm_id,
    )


def get_resources(proxmox: ProxmoxAPI):
    resources = {}

    resources["template"] = select_template(proxmox) or {}

    resources["volume"] = select_storage(proxmox, host=resources["template"]["node"]) or {}

    resources["network"] = select_network(proxmox, host=resources["template"]["node"]) or {}

    try:
        resources["network"]["bridge"] = select_network_bridge(proxmox, host=resources["network"]["node"]) or {}
    except KeyError as error:
        print(error)  # no network selected

    return resources


def delete_vms(proxmox: ProxmoxAPI):
    vms_to_delete = [x for x in select_vms(proxmox)]
    if vms_to_delete and len(vms_to_delete) > 0:
        [print(vm_name(x)) for x in vms_to_delete]
        if confirm("Are you sure to delete these VMs", default=False).ask():
            threads = list()
            for vm in vms_to_delete:
                task = ReturningThread(
                    target=delete_vm,
                    kwargs={
                        "proxmox": proxmox,
                        "vm": vm,
                    },
                )
                threads.append(task)
            start_tasks_delayed(threads, 0)
            return wait_tasks(threads)
        else:
            print("Ok, backing off")
    else:
        print("No VM selected")
    return None


def delete_vm(proxmox: ProxmoxAPI, vm):
    # capture ip address before delete
    vm_ip = None
    try:
        interfaces = proxmox.nodes(vm["node"])(vm["id"]).agent("network-get-interfaces").get()
        if interfaces:
            interface = [int["ip-addresses"] for int in interfaces["result"] if int["name"] == "eth0"][0]
            vm_ip = [addr["ip-address"] for addr in interface if addr["ip-address-type"] == "ipv4"][0]
    except IndexError:
        print(f"Cannot find IP for {vm['name']}")

    if vm["status"] == "running":
        task_waitfor(
            title=f"Stopping {vm['name']}",
            connection=proxmox,
            node=vm["node"],
            task_id=proxmox.nodes(vm["node"])(vm["id"]).status.stop.post(timeout=30),
        )
    task_waitfor(
        title=f"Deleting {vm['name']}",
        connection=proxmox,
        node=vm["node"],
        task_id=proxmox.nodes(vm["node"])(vm["id"]).delete(purge=1),
    )
    if vm_ip:
        # remove from known_hosts
        subprocess.Popen(
            ["ssh-keygen", "-R", vm_ip],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        subprocess.Popen(
            ["ssh-keygen", "-R", vm["name"]],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

        # # remove from dns
        # if vm_ip and config.get()["NETWORK"]["dns_api"] != "":
        #     result = technitium_delete_record(hostname=vm["name"], ip=vm_ip)
        #     if isinstance(result, Response) and result.json()["status"] == "ok":
        #         print(f"dns record deleted for {vm['name']}")

    return vm["name"]


def delete_templates(proxmox: ProxmoxAPI):
    vms_to_delete = [x for x in select_templates(proxmox)]
    if vms_to_delete and len(vms_to_delete) > 0:
        [print(vm_name(x)) for x in vms_to_delete]
        if confirm("Are you sure to delete these templates", default=False).ask():
            for vm in vms_to_delete:
                if proxmox.nodes(vm["node"])(vm["id"]).delete(purge=1):
                    yield vm["name"]
        else:
            print("Ok, backing off")
    else:
        print("No template selected")
        return None


def get_storage_for(connection, content: str = "", plugin: list = [], title: str = "Storage"):
    content_filter = True if content == "" else lambda x: content in x["content"].split(",")
    plugin_filter = True if plugin == "" else lambda x: x["plugintype"] in plugin
    volumes = [x for x in storage(connection) if content_filter(x) and plugin_filter(x)]  # type: ignore
    if len(volumes) == 0:
        fail(f"No storage available for {content} or {plugin}")
    volume = select_from_list(title, volumes, volume_name, volume_id)
    return volume


def task_waitfor(connection, node, task_id, title):
    task_submitted = task_status(connection, node, task_id)
    if task_submitted["status"] == "stopped":
        fail(task_submitted["exitstatus"])

    print(f"{title}...")
    try:
        while task_status(connection, node, task_id)["status"] != "stopped":
            sleep(3)
    except Exception as error:
        return error

    # task_running = spinner(
    #     title=title,
    #     task=connection.nodes(node).tasks(task_id).status.get,
    #     success=lambda x: x["status"] == "stopped",
    # )

    # if not task_running or isinstance(task_running, Exception):
    #     fail(task_running)

    return True


def download_to_storage(connection, node, storage):
    defaults = config.get()
    images = list(defaults["CLOUDIMG"])
    image_name = select_from_list("Image URL", images, lambda x: x, lambda x: x)
    if image_name:
        image_url = defaults["CLOUDIMG"][image_name]
        if not image_url:
            fail("No image selected for upload")
        # Tricking file extension so API accepts it
        target_filename = f"{os.path.basename(image_url)}.iso"
        task_id = connection.nodes(node).storage(storage)("download-url").post(url=image_url, content="iso", filename=target_filename)

        task_waitfor(connection, node, task_id, "Downloading...")


def create_template(proxmox: ProxmoxAPI):
    defaults = config.get()

    base_image = select_from_list(
        "Base image",
        list(defaults["CLOUDIMG"].keys()),
        lambda x: x,
        lambda x: defaults["CLOUDIMG"][x],
    )

    os_storage = get_storage_for(
        proxmox,
        content="images",
        plugin=["lvmthin", "zfspool", "rbd"],
        title="OS Storage",
    )

    if not os_storage:
        fail("No storage is selected for OS disk")

    else:
        node = os_storage["node"]
        base_image_os = ""
        if base_image:
            for key in defaults["CLOUDIMG"]:
                if base_image in defaults["CLOUDIMG"][key].lower():
                    base_image_os = key
            base_image_full_path = f"/var/lib/vz/template/qemu/{defaults['CLOUDIMG'][base_image_os]}"
            template_name = f"{base_image_os}-template"

            all_vms = vms(proxmox)
            if all_vms:
                num_template = len([vm for vm in all_vms if template_name in vm["name"]]) + 1
                template_name += str(num_template)

                with open(get_ssh_key_file()) as file:
                    ssh_key = file.read()

                new_vm_id = pve_nextid(proxmox)

                task_id = proxmox.nodes(node).qemu.post(
                    agent="enabled=1,fstrim_cloned_disks=1,freeze-fs-on-backup=0",
                    arch="x86_64",
                    bios="ovmf",
                    boot="order=scsi0",
                    cpu="host",
                    description="Ezlab template",
                    cdrom=f"{os_storage['storage']}:cloudinit",
                    machine="q35",
                    memory="2048",
                    name=template_name,
                    ostype="l26",
                    scsi0=f"{os_storage['storage']}:0,import-from={base_image_full_path},backup=0,discard=on,cache=unsafe,iothread=1,replicate=0,ssd=1",
                    scsihw="virtio-scsi-single",
                    tags="ezlab",
                    template=0,
                    vmid=new_vm_id,
                    ciuser=defaults["CLOUDINIT"]["vm_username"],
                    cipassword=defaults["CLOUDINIT"]["vm_password"],
                    nameserver=defaults["NETWORK"]["vm_nameserver"],
                    searchdomain=defaults["NETWORK"]["vm_domain"],
                    ciupgrade=0,
                    sshkeys=quote(ssh_key, safe=""),
                )

                if task_waitfor(
                    connection=proxmox,
                    node=node,
                    task_id=task_id,
                    title=f"clone for {template_name}",
                ):
                    # refresh cloudinit drive
                    proxmox.nodes(node).qemu(new_vm_id).cloudinit.put()

                    if task_waitfor(
                        connection=proxmox,
                        node=node,
                        task_id=proxmox.nodes(node).qemu(new_vm_id).template.post(disk="scsi0"),
                        title="make template",
                    ):
                        yield template_name

                else:
                    fail(f"Failed to clone template {template_name}")


def clone_vms(proxmox: ProxmoxAPI, customisations: dict, nodes: list):
    resources = get_resources(proxmox)
    if not resources["template"]:
        fail("No template selected")

    if not resources["volume"]:
        fail("No volume selected")

    if not resources["network"]:
        fail("No network selected")
    name_index = 0  # index for vm names
    threads = list()
    for ez_node in nodes:
        for idx in range(ez_node["count"]):
            name_index += 1
            task = ReturningThread(
                target=clone_vm,
                kwargs={
                    "proxmox": proxmox,
                    "resources": resources,
                    "customisations": customisations,
                    "ez_node": ez_node,
                    "vm_number": name_index,
                },
            )
            threads.append(task)

    start_tasks_delayed(threads, 2)

    return wait_tasks(threads)


def clone_vm(
    proxmox: ProxmoxAPI,
    resources: dict,
    customisations: dict,
    ez_node: dict,
    vm_number: int,
):
    node = resources["template"]["node"]
    template_type = resources["template"]["type"]
    template_id = resources["template"]["vmid"]
    network_bridge_name = resources["network"]["bridge"]["iface"]

    defaults = config.get()
    vm_gateway = defaults["NETWORK"]["vm_gateway"]
    vm_network_prefix = defaults["NETWORK"]["vm_cidr"].split("/")[1]

    nextid = pve_nextid(proxmox)
    new_vm_name = customisations["vmname"] + str(vm_number)
    ipconfig = "ip="
    vm_ip = ""
    if customisations["first_vm_ip"] == "dhcp":
        ipconfig += "dhcp"
    else:
        vm_ip = str(ip_address(customisations["first_vm_ip"]) + (vm_number - 1))  # vm number starting at 1, ip index should start with 0
        ipconfig += f"{vm_ip}/{vm_network_prefix}"
        ipconfig += f",gw={vm_gateway}"

    if task_waitfor(
        connection=proxmox,
        node=node,
        task_id=proxmox.nodes(node)(template_type)(template_id).clone.post(
            newid=nextid,
            name=new_vm_name,
            description=ez_node["name"],
        ),
        title=f"{new_vm_name} cloning",
    ):
        new_vm = proxmox.nodes(node)(template_type)(nextid)

        # configure vm
        new_vm.config.post(
            cores=ez_node["cores"],
            memory=ez_node["memGB"] * 1024,
            net0=f"virtio,bridge={network_bridge_name},firewall=0",
            ipconfig0=ipconfig,
            tags=ez_node["name"],
            onboot=1,
            efidisk0=f"{resources['volume']['storage']}:1,efitype=4m,pre-enrolled-keys=1,size=4M",
        )

        # configure disks
        new_vm.resize.put(disk="scsi0", size=f"{ez_node['os_disk_size']}G")
        # add new disks /// assume no_of_disks are 0 or 1 or 2
        if ez_node["no_of_disks"] > 0:
            new_disk = f"{resources['volume']['storage']}:{ez_node['data_disk_size']},backup=0,discard=on,cache=unsafe,iothread=1,replicate=0,ssd=1"
            new_vm.config.post(scsi1=new_disk)
        if ez_node["no_of_disks"] > 1:
            new_disk = f"{resources['volume']['storage']}:{ez_node['data_disk_size']},backup=0,discard=on,cache=unsafe,iothread=1,replicate=0,ssd=1"
            new_vm.config.post(scsi1=new_disk, scsi2=new_disk)

        # start vm
        new_vm.status.start.post()

    # # add to dns
    # if vm_ip and defaults["NETWORK"]["dns_api"] != "":
    #     # TODO: more integrations
    #     result = technitium_add_record(hostname=new_vm_name, ip=vm_ip)
    #     if isinstance(result, Response) and result.json()["status"] == "ok":
    #         print(f"dns record added for {new_vm_name}:{vm_ip}")
    #     else:
    #         print(result)
    # apply customisations to vm

    if not wait_for_ssh(vm_ip):
        fail(f"ssh failed for {vm_ip}")

    print(f"configuring {vm_ip}")
    if prepare_vm(
        vm_name=new_vm_name,
        vm_ip=vm_ip,
        product_code=ez_node["product"],
    ):
        print(f"{new_vm_name} configured")

    # # reboot for changes
    # task_waitfor(
    #     connection=proxmox,
    #     node=node,
    #     task_id=new_vm.status.reboot.post(timeout=60),
    #     title=f"reboot {new_vm_name}",
    # )
    return new_vm_name, ez_node["product"]



import logging
from nicegui import app, ui, run

from ezlab.ezmeral.ezua import deploy, install, precheck
from ezlab.parameters import UA
from ezinfra.vms import prepare
from ezlab.utils import get_doc, get_fqdn


logger = logging.getLogger("ezlab.ui.ua")


async def runall_action():
    return await prepare_action() and await precheck_action() and await install_action() and await deploy_action()


async def prepare_action():
    orchestrator = app.storage.general[UA].get("orchestrator", None)
    controller = app.storage.general[UA].get("controller", None)
    workers = app.storage.general[UA].get("workers", None)

    if orchestrator is None or controller is None or workers is None or len(workers.split(",")) < 3:
        logger.warning("You have some hosts missing, make sure you have all roles filled in!")
        logger.debug(
            "Orchestrator: %s, Controller: %s, Workers: %s",
            orchestrator,
            controller,
            workers,
        )

    logger.info("Starting configuration for UA")
    app.storage.user["busy"] = True

    hosts = [orchestrator, controller]
    hosts.extend(workers.split(","))

    completed = 0

    for host in hosts:
        result = await run.io_bound(
            prepare,
            hostname=get_fqdn(host),
            hostip=host,
            settings=dict(app.storage.general["config"]),
            addhosts=[h for h in hosts if h != host],
            add_to_noproxy=f"{orchestrator},{controller},{workers}",
            prepare_for=UA,
            dryrun=app.storage.general["config"]["dryrun"],
        )
        if app.storage.general["config"]["dryrun"]:
            get_doc(result)
            completed += 1
        elif result:
            logger.info("[ %s ] ready: %s", host, result)
            completed += 1
        else:
            ui.notify(f"[{host}] preparation failed!", type="warning")

    app.storage.user["busy"] = False

    return True if completed == len(hosts) else False


async def precheck_action():
    logger.info("Running prechecks for UA")
    app.storage.user["busy"] = True

    result = await run.io_bound(precheck)
    logger.debug("RESULT: %s", result)

    app.storage.user["busy"] = False

    return result


async def install_action():
    logger.info("Starting UA installation")
    app.storage.user["busy"] = True
    result = await run.io_bound(install)

    logger.debug("RESULT: %s", result)

    app.storage.user["busy"] = False

    return result


async def deploy_action():
    logger.info("Deploying UA Applications")
    app.storage.user["busy"] = True
    result = await run.io_bound(deploy)

    logger.debug("RESULT: %s", result)

    app.storage.user["busy"] = False

    return result



from ipaddress import ip_address, ip_network
from typing import Annotated
from .connect import (
    connect_to_target,
)
from .common import (
    get_customisations,
    select_nodes,
    target_names,
    fail,
    target_classes,
)
from . import config

import typer

from . import pve
from . import vmw

app = typer.Typer(add_completion=True, no_args_is_help=True)


@app.command()
def vm(
    target: Annotated[target_names, typer.Option("--target", "-t", prompt=True)],
    host: Annotated[str, typer.Option("--host", "-h", envvar="EZLAB_HOST", prompt=True)],
    username: Annotated[str, typer.Option("--username", "-u", envvar="EZLAB_USER", prompt=True)],
    password: Annotated[
        str,
        typer.Option("--password", "-p", prompt=True, hide_input=True, envvar="EZLAB_PASSWORD"),
    ],
):
    connection = connect_to_target(target, host, username, password)

    # app settings
    defaults = config.get()
    # What to deploy
    nodes = select_nodes()

    requested_node_count = sum([int(x["count"]) for x in nodes])
    if len(nodes) < 1:
        fail("No nodes selected")
    else:
        print(f"{requested_node_count} VMs will be created")

    # Check resource availability
    # [TODO] Target specific controls required here, such as if there is enough capacity on selected volume, or enough cores in selected cluster etc.

    # Customisation
    customisations = get_customisations()
    if customisations["first_vm_ip"] != "dhcp":
        if ip_address(customisations["first_vm_ip"]) + requested_node_count not in ip_network(defaults["NETWORK"]["vm_cidr"]).hosts():  # type: ignore
            fail(f"Not enough IP available in selected network.\n Available IPs: {ip_network(defaults['NETWORK']['vm_cidr']).hosts()}")

    new_vms = []
    if connection is not None:
        connection_type = connection.__class__.__name__
        if connection_type == target_classes.pve.value:
            new_vms = pve.clone_vms(
                proxmox=connection,  # type: ignore
                customisations=customisations,
                nodes=nodes,
            )
        elif connection_type == target_classes.vmw.value:
            fail("Vmware cloudinit customisation is not working at this release, check if there is a newer release!")
            new_vms = vmw.clone_vms(
                connection=connection,  # type: ignore
                customisations=customisations,
                nodes=nodes,
            )
        else:
            fail("Unknown connection type")
    try:
        for vm, product in new_vms:
            print(f"{vm} is ready for {product}")
    except TypeError:
        fail("Cannot complete vm creation")


@app.command()
def template(
    target: Annotated[target_names, typer.Option("--target", "-t", prompt=True)],
    host: Annotated[str, typer.Option("--host", "-h", envvar="EZLAB_HOST", prompt=True)],
    username: Annotated[str, typer.Option("--username", "-u", envvar="EZLAB_USER", prompt=True)],
    password: Annotated[
        str,
        typer.Option("--password", "-p", prompt=True, hide_input=True, envvar="EZLAB_PASSWORD"),
    ],
):
    connection = connect_to_target(target, host, username, password)

    vms = []
    if connection is not None:
        connection_type = connection.__class__.__name__
        if connection_type == target_classes.pve.value:
            if "root" not in username:
                fail("You must connect with root to create template.")
            vms = pve.create_template(connection)  # type: ignore
        elif connection_type == target_classes.vmw.value:
            vms = vmw.create_template(connection)  # type: ignore
        else:
            fail("Unknown connection type")

    else:
        fail(f"Cannot open connection to {target}")

    for vm in vms:
        print(f"{vm} is ready")



import io
import logging
import random
import socket
import string
from time import sleep
from paramiko import (
    AuthenticationException,
    AutoAddPolicy,
    BadHostKeyException,
    RSAKey,
    SSHClient,
    SSHException,
)

from ezlab.parameters import PVE

logger = logging.getLogger("ezinfra.remote")


def wait_for_ssh(host, username, privatekey):
    client = SSHClient()
    client.set_missing_host_key_policy(AutoAddPolicy)
    sleep_time = 15
    try:
        while True:
            sleep(sleep_time)
            logger.debug("check ssh for %s@%s", username, host)
            client.connect(
                host,
                username=username,
                # key_filename=privatekey,
                pkey=RSAKey.from_private_key(io.StringIO(privatekey)),
                timeout=sleep_time,
            )
            break
    except (
        BadHostKeyException,
        AuthenticationException,
        ConnectionResetError,
        SSHException,
        socket.error,
    ):
        logger.debug("%s: still waiting ssh...", host)
        sleep(sleep_time)
    except Exception as error:
        logger.warning("[ %s ] ssh failed: %s", host, error)
        return False
    finally:
        client.close()

    sleep(15)  # give some time to service as it doesn't come up right away
    return True


def ssh_run_command(
    host: str,
    username: str,
    keyfile: str,
    command: str,
):
    """
    Run commands over ssh using default credentials
    Params:
        <host>              : hostname/IP to SSH
        <username>          : username to authenticate
        <keyfile>           : private key for <username>
        <command>           : command to run
    Returns:
        Generator with stdout. Calling function should process output with a for loop.
        Prints out stderr directly into terminal.

    """
    client = SSHClient()
    client.set_missing_host_key_policy(AutoAddPolicy)
    command_output = None
    # print(f"Try connection to {host}")
    try:

        # logger.debug("REMOTE SSH GOT %s", command)

        client.connect(
            host,
            username=username,
            # key_filename=keyfile,
            # pkey=RSAKey.from_private_key(io.StringIO(keyfile)),
            pkey=RSAKey.from_private_key_file(keyfile),
            timeout=60,
        )

        # if "hostname -f" not in command:
        #     yield f"SSHCMD: {command}"
        _stdin, _stdout, _stderr = client.exec_command(command, get_pty=True)
        command_output = _stdout.read().decode().strip()

        if _stderr.read().decode().strip() != "":
            yield f"[SSH CMD ERROR]: {_stderr.readlines()}"
        if command_output and command_output != "":
            # should return only the command output
            yield command_output
    except Exception as error:
        raise error

    finally:
        client.close()


def sftp_copy_file(host: str, username: str, keyfile: str, source: str, destfile: str):
    try:
        with SSHClient() as ssh:
            ssh.set_missing_host_key_policy(AutoAddPolicy)
            ssh.connect(hostname=host, username=username, pkey=RSAKey.from_private_key_file(keyfile))
            with ssh.open_sftp() as sftp:
                result = sftp.put(source, destfile)
                if result:
                    logger.debug("%s copied", source)
                else:
                    logger.warning("Failed to copy %s: ", source)

    except Exception as error:
        logger.warning("SFTP Exception %s", error)


def sftp_content_to_file(host: str, username: str, keyfile: str, content: str, filepath: str):
    random_eof = "".join(random.choices(string.ascii_uppercase + string.digits, k=10))
    command = f"""cat << {random_eof} | sudo tee {filepath} > /dev/null
{content}
{random_eof}"""

    # logger.debug("REMOTE SFTP %s", filepath)

    return ssh_run_command(
        host=host,
        username=username,
        keyfile=keyfile,
        command=command,
    )



import logging
from ezinfra.proxy import proxy_files
from ezinfra.remote import sftp_content_to_file, ssh_run_command
from ezinfra.repo import get_epel_repo, get_insecure_registry, get_local_repo
from ezlab.parameters import DF, GENERIC, KVM, PVE, SETUP_COMMANDS
from ezinfra import pve
from ezinfra import kvm


logger = logging.getLogger("ezinfra.vms")


def prepare(
    hostname: str,
    hostip: str,
    settings: dict,
    addhosts: list,
    add_to_noproxy: str,
    prepare_for: str,
    dryrun: bool,
):
    """
    Configure VMs for selected product, taking care of pre-requisites
    Params:
        <hostname>      : VM hostname
        <hostip>        : VM IP address (should be reachable)
        <settings>      : Dictinary of configuration from app settings
        <addhosts>      : Other hosts to add to /etc/hosts file, dict{'ip', 'name'}
        <add_to_noproxy>: add to no_proxy (comma separated hosts/IPs)
        <prepare_for>   : EZUA | EZDF | DFAAS | Generic
        <dryrun>        : Do not execute but report

    Returns <bool>     : True in succesful completion, False otherwise. If dryrun, returns dict of job run
    """

    response = {}

    if dryrun:
        response["task"] = (
            f"Prepare {hostname} ({hostip}) for {prepare_for} {'with hosts: ' + ','.join([h for h in addhosts]) if len(addhosts) > 0 else 'as single node'}"
        )
        response["settings"] = settings

    domain = settings["domain"]
    cidr = settings["cidr"]
    username = settings["username"]
    # password = settings["password"]
    keyfile = settings["privatekeyfile"]
    # commands to execute, always prepend with generic/common commands, such as enable password auth, disable subscription mgr etc.
    commands = list(SETUP_COMMANDS[GENERIC])
    if prepare_for != GENERIC:
        commands += list(SETUP_COMMANDS[prepare_for])
    after_generic_commands = len(list(SETUP_COMMANDS[GENERIC]))

    files = {}
    # "/etc/cloud/templates/hosts.redhat.tmpl": f"127.0.0.1 localhost.localdomain localhost\n{hostip} {hostname}.{domain} {hostname}\n"

    # ensure shortname (no fqdn)
    hostname = hostname.split(".")[0]
    hosts = [
        "127.0.0.1 localhost.localdomain localhost",
        f"{hostip} {hostname}.{domain} {hostname}",
    ]
    # + [f"{socket.gethostbyname(h)} {h}" for h in addhosts]

    files = {"/etc/hosts": "\n".join(hosts)}

    # update env files for proxy
    noproxy = f"{hostip},{add_to_noproxy}" if add_to_noproxy is not None else hostip

    if settings.get("proxy", "").strip() != "":
        for file in proxy_files(settings["proxy"], domain, cidr, noproxy):
            files.update(file)

    # TODO: Not tested, enable insecure registry
    if settings.get("airgap_registry", "").strip() != "" and settings["airgap_registry"].split("://")[0] == "http":
        files.update({"/etc/docker/daemon.json": get_insecure_registry(settings["airgap_registry"])})

    repo_content = ""
    if settings.get("yumrepo", "").strip() != "":
        for repo in ["BaseOS", "AppStream", "PowerTools", "extras"]:
            repo_content += get_local_repo(repo, settings.get("yumrepo")) + "\n"

    else:
        logger.debug("Using system repositories")

    if settings.get("epelrepo", "").strip() != "":
        repo_content += get_epel_repo(settings.get("epelrepo"))

    else:
        epel_default = "sudo subscription-manager repos --enable codeready-builder-for-rhel-8-x86_64-rpms; sudo dnf install -y https://dl.fedoraproject.org/pub/epel/epel-release-latest-8.noarch.rpm"
        commands.insert(after_generic_commands, epel_default)

    if repo_content != "":  # yum or epel repo given
        files.update({"/etc/yum.repos.d/local.repo": repo_content})

    if dryrun:
        response["files"] = files
    else:
        try:
            for filepath, content in files.items():
                logger.info("[ %s ] COPY %s", hostname, filepath)
                for result in sftp_content_to_file(
                    host=hostip,
                    username=username,
                    keyfile=keyfile,
                    content=content,
                    filepath=filepath,
                ):
                    logger.debug("[ %s ] COPYOUT: %s", hostname, result)

        except Exception as error:
            logger.warning("SFTP FAIL: %s", error)
            return False

    if settings.get("yumrepo", "").strip() != "" or settings.get("epelrepo", "").strip() != "":
        commands.insert(
            after_generic_commands,
            "sudo dnf config-manager --set-disabled baseos appstream extras; sudo dnf config-manager --set-disabled 'rhel-8-*'",
        )

    if settings.get("proxy", "").strip() != "":
        commands.insert(
            after_generic_commands + 1,
            "sudo sed -i '/proxy=/d' /etc/dnf/dnf.conf > /dev/null; echo proxy={proxy_line} | sudo tee -a /etc/dnf/dnf.conf >/dev/null".format(
                proxy_line=(settings["proxy"] if settings["proxy"][-1] == "/" else settings["proxy"] + "/")
            ),
        )

    # set hostname for fqdn - required for DF
    if prepare_for == DF:
        commands.insert(after_generic_commands + 2, f"sudo hostnamectl set-hostname {hostname}.{domain}")

    if dryrun:
        response["commands"] = commands

    else:
        try:
            for command in commands:
                logger.info("[ %s ] SSH: %s", hostname, command)
                for output in ssh_run_command(host=hostip, username=username, keyfile=keyfile, command=command):
                    logger.debug("[ %s ] %s", hostname, output)
        except Exception as error:
            logger.warning("SSH FAIL %s", error)
            return False

    return response if dryrun else True


def clone(
    target: str,
    resources: set,
    settings: dict,
    # vm_number: int,
    vm_count: int,
    dryrun: bool,
):

    if target == PVE:
        return pve.clone(
            resources=resources,
            settings=settings,
            # vm_number=vm_number,
            vm_count=vm_count,
            dryrun=dryrun,
        )
    elif target == KVM:
        return kvm.clone(
            resources=resources,
            settings=settings,
            # vm_number=vm_number,
            vm_count=vm_count,
            dryrun=dryrun,
        )
    else:
        logger.debug("Unknown target %s", target)
        return False



from paramiko import AutoAddPolicy, SSHClient
from ezinfra.remote import sftp_content_to_file, ssh_run_command
from ezinfra.repo import get_mapr_repo
from ezlab.utils import *
from ezlab.parameters import *
from paramiko_expect import SSHClientInteraction


logger = logging.getLogger("ezdf")


def cluster_install(
    hosts: list[str],
    settings: dict,  # settings from "config" key
    config: dict,  # settings from "EZDF" key
    dryrun: bool = True,
):
    hostnames = []
    name = config.get("cluster_name", "ezlab")
    disks = config.get("maprdisks", "/dev/sdz")

    response = {}
    if dryrun:
        response["task"] = f"Create Data Fabric cluster {name} on {','.join(hosts)} using disk(s) {disks}"
        response["settings"] = settings

    # lazy way to catch exceptions
    try:
        username = settings.get("username")
        # used for connecting to nodes from this app
        keyfile = settings.get("privatekeyfile")
        # used for connection between cluster nodes
        password = settings.get("password")

        # get fqdn for each host
        if dryrun:

            def get_hostname_command(host):
                return f"ssh {settings.get('username')}@{host} 'hostname -f'"

            response["commands"] = [get_hostname_command(host) for host in hosts]

        else:
            for host in hosts:
                for result in ssh_run_command(
                    host=host.strip(),
                    username=username,
                    keyfile=keyfile,
                    command="hostname -f",
                ):
                    hostnames.append(result)

        commands: list[str] = INSTALL_COMMANDS[DF]

        if config.get("maprrepoislocal", False):
            # using local repository for mapr packages
            ezrepo: str = config.get("maprlocalrepo", "").rstrip("/")
            repouser = config["maprlocalrepousername"]
            repopass = config["maprlocalrepopassword"]

            wgetauthstring = f"--user={repouser} --password={repopass}" if config.get("maprlocalrepoauth", False) else ""

            proto, path, *_ = ezrepo.split("://")
            repourl = f"{proto}://{repouser}:{repopass}@{path}" if config.get("maprlocalrepoauth", False) else ezrepo

            commands.insert(
                0,
                f"[ -f /tmp/mapr-setup.sh ] || wget -q {wgetauthstring} {ezrepo.rstrip('/')}/installer/redhat/mapr-setup.sh -P /tmp; chmod +x /tmp/mapr-setup.sh",
            )

            commands.insert(
                1,
                f"curl -s -k https://127.0.0.1:9443/ > /dev/null || (sudo /tmp/mapr-setup.sh -y -r {repourl.rstrip('/')} && echo 'Sleeping to let installer to be ready' && sleep 30)",
            )

        else:
            # using HPE repository
            ezrepo = f"https://package.ezmeral.hpe.com/releases"
            repouser = config["maprrepouser"]
            repotoken = config["maprrepotoken"]

            commands.insert(
                0,
                f"[ -f /tmp/mapr-setup.sh ] || wget -q --user={repouser} --password= {repotoken} {ezrepo.rstrip('/')}/installer/redhat/mapr-setup.sh -P /tmp; chmod +x /tmp/mapr-setup.sh",
            )

            commands.insert(
                1,
                f"curl -s -k https://127.0.0.1:9443/ > /dev/null || (sudo /tmp/mapr-setup.sh -y --repo-user {repouser} --repo-token {repotoken} && echo 'Sleeping to let installer to be ready' && sleep 30)",
            )

        first_host = hosts[0].strip()

        # using this to provide UI link to the installer
        commands.insert(2, f"echo 'Installer ready at https://{first_host}:9443/'")

        # Create and upload stanza file
        stanza = get_mapr_stanza(
            hosts=hostnames,
            username=username,
            password=password,
            disks=disks.split(","),
            cluster_name=name,
        )

        if dryrun:
            stanza = get_mapr_stanza(
                hosts=hosts,
                username=username,
                password=password,
                disks=disks.split(","),
                cluster_name=name,
            )
            response["files"] = {"/tmp/mapr.stanza": stanza}

        else:
            logger.info("[ %s ] COPY /tmp/mapr.stanza", first_host)
            # print(f"[ {first_host} ] COPY /tmp/mapr.stanza")
            for result in sftp_content_to_file(
                host=first_host,
                username=username,
                keyfile=keyfile,
                content=stanza,
                filepath="/tmp/mapr.stanza",
            ):
                logger.debug("[ %s ]: %s", first_host, result)

            logger.info("[ %s ] starting installer", first_host)

        if dryrun:
            response["commands"].extend(commands)

        else:
            for command in commands:
                logger.info("[ %s ] RUN %s", first_host, command)
                for result in ssh_run_command(
                    host=first_host,
                    username=username,
                    keyfile=keyfile,
                    command=command,
                ):
                    logger.debug("[ %s ] %s", first_host, result)
                    if "Installer ready at" in result:
                        APP_STATUS_CONTAINER.append((f"{name.upper()} Installer", f"https://mapr:mapr@{first_host}:9443/"))
                        status_container.refresh()

            logger.info(
                "Cluster ready at https://%s:8443/",
                first_host,
            )
            APP_STATUS_CONTAINER.append((f"{name.upper()} MCS", f"https://mapr:mapr@{first_host}:8443/app/mcs/"))
            status_container.refresh()

    except Exception as error:
        logger.warning("Cluster creation exception: %s", error)
        if dryrun:
            return response
        else:
            return False

    if dryrun:
        return response
    else:
        return True


def client_install(
    username: str,
    keyfile: str,
    config: dict,  # settings from "EZDF" key
):
    server = config.get("connect_to")
    client = config.get("maprclient")
    cluster_name = config.get("cluster_name")

    if config.get("maprrepoislocal", False):
        # using local repository for mapr packages
        ezrepo: str = config.get("maprlocalrepo", "").rstrip("/")
    else:
        # using HPE repository
        ezrepo = f"https://package.ezmeral.hpe.com/releases"

    repo_content = get_mapr_repo(ezrepo)
    repofilename = "/etc/yum.repos.d/mapr.repo"

    logger.info("[ %s ] Install client for %s to %s", client, cluster_name, server)

    if client is not None:
        try:
            for out in sftp_content_to_file(
                host=client,
                username=username,
                keyfile=keyfile,
                content=repo_content,
                filepath=repofilename,
            ):
                logger.debug(f"[ {client} ] COPY {repofilename}: {out}")

        except Exception as error:
            logger.warning("Client file copy exception: %s", error)
            return False

    else:  # working on local machine
        with open(repofilename, "w") as repofile:
            repofile.write(repo_content)

    commands = INSTALL_COMMANDS[DFCLIENT]

    for file in DF_SECURE_FILES:
        commands.append(f"sudo curl -s --insecure --user mapr:mapr sftp://{server}{file} --output {file}")

    # Run these last
    # Configure client for cluster
    commands.append(f"sudo /opt/mapr/server/configure.sh -c -C {server} -N {cluster_name} -secure")
    # Generate ticket for root and enable posix-client (/mapr mount)
    commands.append(
        """echo mapr | sudo maprlogin password -user mapr ;
        sudo maprlogin generateticket -user mapr -type service -out /opt/mapr/conf/maprfuseticket -duration 30:0:0 -renewal 90:0:0;
        [ -d /mapr ] || sudo mkdir /mapr;
        sudo systemctl start mapr-posix-client-basic.service
        """
    )
    for command in commands:
        # print(f"Runnning {command}")
        logger.info("[ %s ] RUN %s", client, command)
        try:
            if client is not None:
                for result in ssh_run_command(host=client, username=username, keyfile=keyfile, command=command):
                    logger.debug(f"[ %s ] $s", client, result)

            else:  # running locally
                result = subprocess.run(
                    command.split(" "), shell=True, capture_output=True, text=True, check=True, universal_newlines=True, timeout=300
                )
                logger.info("[ LOCALHOST ] %s: %s", command, result)

        except Exception as error:
            logger.warning("[ %s ] CMD %s failed with: %s", client, command, error)
            return False

    return True


def xcluster(
    config: dict,
    settings: dict,
    dryrun: bool = True,
):
    local = settings["crosslocalcldb"]
    remote = settings["crossremotecldb"]
    adminuser = settings["crossadminuser"]
    adminpassword = settings["crossadminpassword"]
    username = config["username"]
    keyfile = config["privatekeyfile"]

    logger.info("Create cross cluster connectivity between %s and %s", local, remote)

    response = {}
    if dryrun:
        response["task"] = f"Cross-cluster setup between {local} and {remote}"
        response["settings"] = {**config, **settings}

    else:  # get truststore passwords
        try:
            local_truststore_password = ""
            remote_truststore_password = ""

            for out in ssh_run_command(
                host=local,
                command="sudo grep ssl.server.truststore /opt/mapr/conf/store-passwords.txt | cut -d'=' -f2",
                username=username,
                keyfile=keyfile,
            ):
                local_truststore_password = out

            for out in ssh_run_command(
                host=remote,
                command="sudo grep ssl.server.truststore /opt/mapr/conf/store-passwords.txt | cut -d'=' -f2",
                username=username,
                keyfile=keyfile,
            ):
                remote_truststore_password = out

            if local_truststore_password and remote_truststore_password:
                logger.info(f"using local truststore password: {local_truststore_password}")
                logger.info(f"using remote truststore password: {remote_truststore_password}")
            else:
                logger.warning(f"ERROR: failed to get truststore passwords, local: {local_truststore_password}, remote: {remote_truststore_password}")
                return False
        except Exception as error:
            logger.warning("Cross cluster setup exception: %s", error)
            return False

    cross_cluster_setup = f"sudo rm -rf /tmp/mapr-xcs; sudo -i -u {adminuser} /opt/mapr/server/configure-crosscluster.sh create all \
    -localcrossclusteruser {adminuser} -remotecrossclusteruser {adminuser} \
    -localtruststorepassword {local_truststore_password if local_truststore_password else 'LOCAL_TRUSTSTORE_PASSWORD'} \
    -remotetruststorepassword {remote_truststore_password if remote_truststore_password else 'REMOTE_TRUSTSTORE_PASSWORD'} \
    -remoteip {remote} -localuser {adminuser} -remoteuser {adminuser}"

    if dryrun:
        response["commands"] = f"ssh {local} sudo grep ssl.server.truststore /opt/mapr/conf/store-passwords.txt | cut -d'=' -f2"
        response["commands"] = f"ssh {remote} sudo grep ssl.server.truststore /opt/mapr/conf/store-passwords.txt | cut -d'=' -f2"
        response["commands"] = cross_cluster_setup
        return response

    else:
        try:
            logger.info("If fails run the configure-crosscluster.sh command on one of the cluster nodes:")
            logger.info(cross_cluster_setup)

            client = SSHClient()
            client.set_missing_host_key_policy(AutoAddPolicy)

            client.connect(hostname=local, username=username, key_filename=keyfile)

            interact = SSHClientInteraction(client, timeout=300, display=True)
            prompt = f".*[{username}@{local} ~]$.*"
            interact.expect(prompt)
            # print(interact.current_output_clean)

            interact.send("sudo dnf -q install -y pssh expect")
            interact.expect(prompt)

            interact.send(cross_cluster_setup)

            interact.expect("Enter password for mapr user.*")
            interact.send(adminpassword)
            interact.expect("Enter password for mapr user.*")
            interact.send(adminpassword)

            interact.expect(prompt)
            interact.send("exit")
            interact.expect()
        except Exception as error:
            logger.warning("Cross cluster setup exception: %s", error)
            return False

        logger.info("crosscluster-setup.sh finished, check logs for errors")
        return True


def get_mapr_stanza(hosts: list, username: str, password: str, disks: list, cluster_name: str):
    """
    Return stanza for Data Fabric installation
    """
    hosts_yaml = ""
    for host in hosts:
        hosts_yaml += f"  -   {host}\n"

    disks_yaml = ""
    for disk in disks:
        disks_yaml += f"  -   {disk}\n"

    return f"""
environment:
  mapr_core_version: {MAPR_CORE_VERSION}
config:
  hosts:
{hosts_yaml}
  ssh_id: {username}
  ssh_password: {password}
  enable_nfs: False
  db_admin_user: root
  db_admin_password: mapr123
  log_admin_password: mapr123
  metrics_ui_admin_password: mapr123
  enable_encryption_at_rest: True
  license_type: M7
  mep_version: {MAPR_MEP_VERSION}
  disks:
{disks_yaml}
  disk_format: true
  disk_stripe: 1
  cluster_name: {cluster_name}
  services:
    template-05-converged:
    # template-30-maprdb3:
    # template-20-drill:
    mapr-keycloak:
      enabled: False
    mapr-hivemetastore:
      database:
        name: hive
        user: hive
        password: mapr123
    # this conflicts with keycloak
    mapr-hbase-rest:
      enabled: False
    mapr-grafana:
      enabled: True
    mapr-opentsdb:
      enabled: True
    mapr-collectd:
      enabled: True
    mapr-fluentd:
      enabled: False
    mapr-kibana:
      enabled: False
    mapr-elasticsearch:
      enabled: False
    mapr-data-access-gateway:
    mapr-mastgateway:
    mapr-streams:
    mapr-airflow:
    mapr-hue:
      enabled: False
    mapr-httpfs:
      enabled: False
"""
